#!/usr/bin/env python
"""
_WMBSHelper_t_

Unit tests for the WMBSHelper class.
"""

from builtins import next
import os
import threading
import unittest

from WMCore.BossAir.BossAirAPI import BossAirAPI
from WMCore.Configuration import loadConfigurationFile
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Mask import Mask
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMCore.Services.DBS.DBSReader import DBSReader
from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMBase import getTestBase
from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMCore.WMSpec.StdSpecs.TaskChain import TaskChainWorkloadFactory
from WMCore.WMSpec.WMWorkload import WMWorkload, WMWorkloadHelper
from WMCore.WorkQueue.WMBSHelper import WMBSHelper
from WMCore.WorkQueue.WMBSHelper import killWorkflow
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.Emulators.WMSpecGenerator.Samples.BasicProductionWorkload import getProdArgs, taskChainWorkload
from WMQuality.Emulators.WMSpecGenerator.WMSpecGenerator import createConfig
from WMQuality.TestInitCouchApp import TestInitCouchApp

rerecoArgs = ReRecoWorkloadFactory.getTestArguments()
mcArgs = getProdArgs()

BLOCK1 = '03fe83c2-0c23-11e1-b764-003048caaace'
BLOCK2 = '04be2fcc-0b8f-11e1-b764-003048caaace'


def getFirstTask(wmspec):
    """Return the 1st top level task"""
    # http://www.logilab.org/ticket/8774
    # pylint: disable=E1101,E1103
    return next(wmspec.taskIterator())


class WMBSHelperTest(EmulatedUnitTestCase):
    def setUp(self):
        """
        _setUp_

        """
        super(WMBSHelperTest, self).setUp()

        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)
        self.testInit.setupCouch("wmbshelper_t/jobs", "JobDump")
        self.testInit.setupCouch("wmbshelper_t/fwjrs", "FWJRDump")
        self.testInit.setupCouch("config_test", "GroupUser", "ConfigCache")
        os.environ["COUCHDB"] = "wmbshelper_t"
        self.testInit.setSchema(customModules=["WMCore.WMBS",
                                               "WMComponent.DBS3Buffer",
                                               "WMCore.BossAir",
                                               "WMCore.ResourceControl"],
                                useDefault=False)

        self.workDir = self.testInit.generateWorkDir()

        self.wmspec = self.createWMSpec()
        self.topLevelTask = getFirstTask(self.wmspec)
        self.inputDataset = self.topLevelTask.inputDataset()
        self.dataset = self.topLevelTask.getInputDatasetPath()
        self.dbs = DBSReader(self.inputDataset.dbsurl)
        self.rucioAcct = "wmcore_transferor"
        self.rucio = Rucio(self.rucioAcct)
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=threading.currentThread().logger,
                                     dbinterface=threading.currentThread().dbi)

        self.configFile = EmulatorSetup.setupWMAgentConfig()
        self.config = loadConfigurationFile(self.configFile)

        self.config.component_("JobSubmitter")
        self.config.JobSubmitter.submitDir = self.workDir
        self.config.JobSubmitter.submitScript = os.path.join(getTestBase(),
                                                             'WMComponent_t/JobSubmitter_t',
                                                             'submit.sh')

        return

    def tearDown(self):
        """
        _tearDown_

        Clear out the database.
        """
        self.testInit.clearDatabase()
        self.testInit.tearDownCouch()
        self.testInit.delWorkDir()
        EmulatorSetup.deleteConfig(self.configFile)
        super(WMBSHelperTest, self).tearDown()

        return

    def setupForKillTest(self, baAPI=None):
        """
        _setupForKillTest_

        Inject a workflow into WMBS that has a processing task, a merge task and
        a cleanup task.  Inject files into the various tasks at various
        processing states (acquired, complete, available...).  Also create jobs
        for each subscription in various states.
        """
        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        dummyLocationAction = daoFactory(classname="Locations.New")
        changeStateAction = daoFactory(classname="Jobs.ChangeState")
        resourceControl = ResourceControl()
        resourceControl.insertSite(siteName='site1', pnn='goodse.cern.ch',
                                   ceName='site1', plugin="TestPlugin")
        resourceControl.insertThreshold(siteName='site1', taskType='Processing', \
                                        maxSlots=10000, pendingSlots=10000)

        userDN = 'someDN'
        userAction = daoFactory(classname="Users.New")
        userAction.execute(dn=userDN, group_name='DEFAULT', role_name='DEFAULT')

        inputFileset = Fileset("input")
        inputFileset.create()

        inputFileA = File("lfnA", locations="goodse.cern.ch")
        inputFileB = File("lfnB", locations="goodse.cern.ch")
        inputFileC = File("lfnC", locations="goodse.cern.ch")
        inputFileA.create()
        inputFileB.create()
        inputFileC.create()

        inputFileset.addFile(inputFileA)
        inputFileset.addFile(inputFileB)
        inputFileset.addFile(inputFileC)
        inputFileset.commit()

        unmergedOutputFileset = Fileset("unmerged")
        unmergedOutputFileset.create()

        unmergedFileA = File("ulfnA", locations="goodse.cern.ch")
        unmergedFileB = File("ulfnB", locations="goodse.cern.ch")
        unmergedFileC = File("ulfnC", locations="goodse.cern.ch")
        unmergedFileA.create()
        unmergedFileB.create()
        unmergedFileC.create()

        unmergedOutputFileset.addFile(unmergedFileA)
        unmergedOutputFileset.addFile(unmergedFileB)
        unmergedOutputFileset.addFile(unmergedFileC)
        unmergedOutputFileset.commit()

        mainProcWorkflow = Workflow(spec="spec1", owner="Steve",
                                    name="Main", task="Proc")
        mainProcWorkflow.create()
        mainProcMergeWorkflow = Workflow(spec="spec1", owner="Steve",
                                         name="Main", task="ProcMerge")
        mainProcMergeWorkflow.create()
        mainCleanupWorkflow = Workflow(spec="spec1", owner="Steve",
                                       name="Main", task="Cleanup")
        mainCleanupWorkflow.create()

        self.mainProcSub = Subscription(fileset=inputFileset,
                                        workflow=mainProcWorkflow,
                                        type="Processing")
        self.mainProcSub.create()
        self.mainProcSub.acquireFiles(inputFileA)
        self.mainProcSub.completeFiles(inputFileB)

        procJobGroup = JobGroup(subscription=self.mainProcSub)
        procJobGroup.create()
        self.procJobA = Job(name="ProcJobA")
        self.procJobA["state"] = "new"
        self.procJobA["location"] = "site1"
        self.procJobB = Job(name="ProcJobB")
        self.procJobB["state"] = "executing"
        self.procJobB["location"] = "site1"
        self.procJobC = Job(name="ProcJobC")
        self.procJobC["state"] = "complete"
        self.procJobC["location"] = "site1"
        self.procJobA.create(procJobGroup)
        self.procJobB.create(procJobGroup)
        self.procJobC.create(procJobGroup)

        self.mainMergeSub = Subscription(fileset=unmergedOutputFileset,
                                         workflow=mainProcMergeWorkflow,
                                         type="Merge")
        self.mainMergeSub.create()
        self.mainMergeSub.acquireFiles(unmergedFileA)
        self.mainMergeSub.failFiles(unmergedFileB)

        mergeJobGroup = JobGroup(subscription=self.mainMergeSub)
        mergeJobGroup.create()
        self.mergeJobA = Job(name="MergeJobA")
        self.mergeJobA["state"] = "exhausted"
        self.mergeJobA["location"] = "site1"
        self.mergeJobB = Job(name="MergeJobB")
        self.mergeJobB["state"] = "cleanout"
        self.mergeJobB["location"] = "site1"
        self.mergeJobC = Job(name="MergeJobC")
        self.mergeJobC["state"] = "new"
        self.mergeJobC["location"] = "site1"
        self.mergeJobA.create(mergeJobGroup)
        self.mergeJobB.create(mergeJobGroup)
        self.mergeJobC.create(mergeJobGroup)

        self.mainCleanupSub = Subscription(fileset=unmergedOutputFileset,
                                           workflow=mainCleanupWorkflow,
                                           type="Cleanup")
        self.mainCleanupSub.create()
        self.mainCleanupSub.acquireFiles(unmergedFileA)
        self.mainCleanupSub.completeFiles(unmergedFileB)

        cleanupJobGroup = JobGroup(subscription=self.mainCleanupSub)
        cleanupJobGroup.create()
        self.cleanupJobA = Job(name="CleanupJobA")
        self.cleanupJobA["state"] = "new"
        self.cleanupJobA["location"] = "site1"
        self.cleanupJobB = Job(name="CleanupJobB")
        self.cleanupJobB["state"] = "executing"
        self.cleanupJobB["location"] = "site1"
        self.cleanupJobC = Job(name="CleanupJobC")
        self.cleanupJobC["state"] = "complete"
        self.cleanupJobC["location"] = "site1"
        self.cleanupJobA.create(cleanupJobGroup)
        self.cleanupJobB.create(cleanupJobGroup)
        self.cleanupJobC.create(cleanupJobGroup)

        jobList = [self.procJobA, self.procJobB, self.procJobC,
                   self.mergeJobA, self.mergeJobB, self.mergeJobC,
                   self.cleanupJobA, self.cleanupJobB, self.cleanupJobC]

        changeStateAction.execute(jobList)

        if baAPI:
            for job in jobList:
                job['plugin'] = 'TestPlugin'
                job['userdn'] = userDN
                job['usergroup'] = 'DEFAULT'
                job['userrole'] = 'DEFAULT'
                job['custom']['location'] = 'site1'
            baAPI.createNewJobs(wmbsJobs=jobList)

        # We'll create an unrelated workflow to verify that it isn't affected
        # by the killing code.
        bogusFileset = Fileset("dontkillme")
        bogusFileset.create()

        bogusFileA = File("bogus/lfnA", locations="goodse.cern.ch")
        bogusFileA.create()
        bogusFileset.addFile(bogusFileA)
        bogusFileset.commit()

        bogusWorkflow = Workflow(spec="spec2", owner="Steve",
                                 name="Bogus", task="Proc")
        bogusWorkflow.create()
        self.bogusSub = Subscription(fileset=bogusFileset,
                                     workflow=bogusWorkflow,
                                     type="Processing")
        self.bogusSub.create()
        self.bogusSub.acquireFiles(bogusFileA)
        return

    def verifyFileKillStatus(self):
        """
        _verifyFileKillStatus_

        Verify that all files were killed correctly.  The status of files in
        Cleanup and LogCollect subscriptions isn't modified.  Status of
        already completed and failed files is not modified.  Also verify that
        the bogus subscription is untouched.
        """
        failedFiles = self.mainProcSub.filesOfStatus("Failed")
        acquiredFiles = self.mainProcSub.filesOfStatus("Acquired")
        completedFiles = self.mainProcSub.filesOfStatus("Completed")
        availableFiles = self.mainProcSub.filesOfStatus("Available")
        bogusAcquiredFiles = self.bogusSub.filesOfStatus("Acquired")

        self.assertEqual(len(availableFiles), 0, \
                         "Error: There should be no available files.")
        self.assertEqual(len(acquiredFiles), 0, \
                         "Error: There should be no acquired files.")
        self.assertEqual(len(bogusAcquiredFiles), 1, \
                         "Error: There should be one acquired file.")

        self.assertEqual(len(completedFiles), 3, \
                         "Error: There should be only one completed file.")
        goldenLFNs = ["lfnA", "lfnB", "lfnC"]
        for completedFile in completedFiles:
            self.assertTrue(completedFile["lfn"] in goldenLFNs, \
                            "Error: Extra completed file.")
            goldenLFNs.remove(completedFile["lfn"])

        self.assertEqual(len(failedFiles), 0, \
                         "Error: There should be no failed files.")

        self.assertEqual(len(goldenLFNs), 0, \
                         "Error: Missing LFN")

        failedFiles = self.mainMergeSub.filesOfStatus("Failed")
        acquiredFiles = self.mainMergeSub.filesOfStatus("Acquired")
        completedFiles = self.mainMergeSub.filesOfStatus("Completed")
        availableFiles = self.mainMergeSub.filesOfStatus("Available")

        self.assertEqual(len(acquiredFiles), 0, \
                         "Error: Merge subscription should have 0 acq files.")
        self.assertEqual(len(availableFiles), 0, \
                         "Error: Merge subscription should have 0 avail files.")

        self.assertEqual(len(failedFiles), 1, \
                         "Error: Merge subscription should have 1 failed files.")
        self.assertEqual(list(failedFiles)[0]["lfn"], "ulfnB",
                         "Error: Wrong failed file.")

        self.assertEqual(len(completedFiles), 2, \
                         "Error: Merge subscription should have 2 compl files.")
        goldenLFNs = ["ulfnA", "ulfnC"]
        for completedFile in completedFiles:
            self.assertTrue(completedFile["lfn"] in goldenLFNs, \
                            "Error: Extra complete file.")
            goldenLFNs.remove(completedFile["lfn"])

        self.assertEqual(len(goldenLFNs), 0, \
                         "Error: Missing LFN")

        failedFiles = self.mainCleanupSub.filesOfStatus("Failed")
        acquiredFiles = self.mainCleanupSub.filesOfStatus("Acquired")
        completedFiles = self.mainCleanupSub.filesOfStatus("Completed")
        availableFiles = self.mainCleanupSub.filesOfStatus("Available")

        self.assertEqual(len(failedFiles), 0, \
                         "Error: Cleanup subscription should have 0 fai files.")

        self.assertEqual(len(acquiredFiles), 1, \
                         "Error: There should be only one acquired file.")
        self.assertEqual(list(acquiredFiles)[0]["lfn"], "ulfnA", \
                         "Error: Wrong acquired LFN.")

        self.assertEqual(len(completedFiles), 1, \
                         "Error: There should be only one completed file.")
        self.assertEqual(list(completedFiles)[0]["lfn"], "ulfnB", \
                         "Error: Wrong completed LFN.")

        self.assertEqual(len(availableFiles), 1, \
                         "Error: There should be only one available file.")
        self.assertEqual(list(availableFiles)[0]["lfn"], "ulfnC", \
                         "Error: Wrong completed LFN.")

        return

    def verifyJobKillStatus(self):
        """
        _verifyJobKillStatus_

        Verify that jobs are killed correctly.  Jobs belonging to Cleanup and
        LogCollect subscriptions are not killed.  The status of jobs that have
        already finished running is not changed.
        """
        self.procJobA.load()
        self.procJobB.load()
        self.procJobC.load()

        self.assertEqual(self.procJobA["state"], "killed", \
                         "Error: Proc job A should be killed.")
        self.assertEqual(self.procJobB["state"], "killed", \
                         "Error: Proc job B should be killed.")
        self.assertEqual(self.procJobC["state"], "complete", \
                         "Error: Proc job C should be complete.")

        self.mergeJobA.load()
        self.mergeJobB.load()
        self.mergeJobC.load()

        self.assertEqual(self.mergeJobA["state"], "exhausted", \
                         "Error: Merge job A should be exhausted.")
        self.assertEqual(self.mergeJobB["state"], "cleanout", \
                         "Error: Merge job B should be cleanout.")
        self.assertEqual(self.mergeJobC["state"], "killed", \
                         "Error: Merge job C should be killed.")

        self.cleanupJobA.load()
        self.cleanupJobB.load()
        self.cleanupJobC.load()

        self.assertEqual(self.cleanupJobA["state"], "new", \
                         "Error: Cleanup job A should be new.")
        self.assertEqual(self.cleanupJobB["state"], "executing", \
                         "Error: Cleanup job B should be executing.")
        self.assertEqual(self.cleanupJobC["state"], "complete", \
                         "Error: Cleanup job C should be complete.")
        return

    def createTestWMSpec(self):
        """
        _createTestWMSpec_

        Create a WMSpec that has a processing, merge, cleanup and skims tasks that
        can be used by the subscription creation test.
        """
        testWorkload = WMWorkloadHelper(WMWorkload("TestWorkload"))
        testWorkload.setDashboardActivity("TestReReco")
        testWorkload.setSpecUrl("/path/to/workload")
        testWorkload.setOwnerDetails("sfoulkes", "DMWM", {'dn': 'MyDN'})

        procTask = testWorkload.newTask("ProcessingTask")
        procTask.setTaskType("Processing")
        procTask.setSplittingAlgorithm("FileBased", files_per_job=1)
        procTaskCMSSW = procTask.makeStep("cmsRun1")
        procTaskCMSSW.setStepType("CMSSW")
        procTaskCMSSWHelper = procTaskCMSSW.getTypeHelper()
        procTask.setTaskType("Processing")
        procTask.setSiteWhitelist(["site1"])
        procTask.setSiteBlacklist(["site2"])
        procTask.applyTemplates()

        procTaskCMSSWHelper.addOutputModule("OutputA",
                                            primaryDataset="bogusPrimary",
                                            processedDataset="bogusProcessed",
                                            dataTier="DataTierA",
                                            lfnBase="bogusUnmerged",
                                            mergedLFNBase="bogusMerged",
                                            filterName=None)

        mergeTask = procTask.addTask("MergeTask")
        mergeTask.setInputReference(procTaskCMSSW, outputModule="OutputA", dataTier='DataTierA')
        mergeTask.setTaskType("Merge")
        mergeTask.setSplittingAlgorithm("WMBSMergeBySize", min_merge_size=1,
                                        max_merge_size=2, max_merge_events=3)
        mergeTaskCMSSW = mergeTask.makeStep("cmsRun1")
        mergeTaskCMSSW.setStepType("CMSSW")
        mergeTaskCMSSWHelper = mergeTaskCMSSW.getTypeHelper()
        mergeTask.setTaskType("Merge")
        mergeTask.applyTemplates()

        mergeTaskCMSSWHelper.addOutputModule("Merged",
                                             primaryDataset="bogusPrimary",
                                             processedDataset="bogusProcessed",
                                             dataTier="DataTierA",
                                             lfnBase="bogusUnmerged",
                                             mergedLFNBase="bogusMerged",
                                             filterName=None)

        cleanupTask = procTask.addTask("CleanupTask")
        cleanupTask.setInputReference(procTaskCMSSW, outputModule="OutputA", dataTier="DataTierA")
        cleanupTask.setTaskType("Merge")
        cleanupTask.setSplittingAlgorithm("SiblingProcessingBased", files_per_job=50)
        cleanupTaskCMSSW = cleanupTask.makeStep("cmsRun1")
        cleanupTaskCMSSW.setStepType("CMSSW")
        cleanupTask.setTaskType("Cleanup")
        cleanupTask.applyTemplates()

        skimTask = mergeTask.addTask("SkimTask")
        skimTask.setTaskType("Skim")
        skimTask.setInputReference(mergeTaskCMSSW, outputModule="Merged", dataTier="DataTierA")
        skimTask.setSplittingAlgorithm("FileBased", files_per_job=1, include_parents=True)
        skimTaskCMSSW = skimTask.makeStep("cmsRun1")
        skimTaskCMSSW.setStepType("CMSSW")
        skimTaskCMSSWHelper = skimTaskCMSSW.getTypeHelper()
        skimTask.setTaskType("Skim")
        skimTask.applyTemplates()

        skimTaskCMSSWHelper.addOutputModule("SkimOutputA",
                                            primaryDataset="bogusPrimary",
                                            processedDataset="bogusProcessed",
                                            dataTier="DataTierA",
                                            lfnBase="bogusUnmerged",
                                            mergedLFNBase="bogusMerged",
                                            filterName=None)

        skimTaskCMSSWHelper.addOutputModule("SkimOutputB",
                                            primaryDataset="bogusPrimary",
                                            processedDataset="bogusProcessed",
                                            dataTier="DataTierB",
                                            lfnBase="bogusUnmerged",
                                            mergedLFNBase="bogusMerged",
                                            filterName=None)

        return testWorkload

    def setupMCWMSpec(self):
        """Setup MC workflow"""
        self.wmspec = self.createMCWMSpec()
        self.topLevelTask = getFirstTask(self.wmspec)
        self.inputDataset = self.topLevelTask.inputDataset()
        self.dataset = self.topLevelTask.getInputDatasetPath()
        self.dbs = None

        # add sites that would normally be added by operator via resource_control
        locationDAO = self.daoFactory(classname="Locations.New")
        self.pnns = []
        for site in ['T2_XX_SiteA', 'T2_XX_SiteB']:
            locationDAO.execute(siteName=site, pnn=site)
            self.pnns.append(site)

    def createWMSpec(self, name='ReRecoWorkload'):
        factory = ReRecoWorkloadFactory()
        rerecoArgs["ConfigCacheID"] = createConfig(rerecoArgs["CouchDBName"])
        wmspec = factory.factoryWorkloadConstruction(name, rerecoArgs)
        wmspec.setSpecUrl("/path/to/workload")
        wmspec.setSubscriptionInformation(custodialSites=[],
                                          nonCustodialSites=[],
                                          priority="Low")
        return wmspec

    def createMCWMSpec(self, name='MonteCarloWorkload'):
        mcArgs = TaskChainWorkloadFactory.getTestArguments()
        mcArgs["CouchDBName"] = rerecoArgs["CouchDBName"]
        mcArgs["Task1"]["ConfigCacheID"] = createConfig(mcArgs["CouchDBName"])

        wmspec = taskChainWorkload(name, mcArgs)
        wmspec.setSpecUrl("/path/to/workload")
        getFirstTask(wmspec).addProduction(totalevents=10000)
        return wmspec

    def getDBS(self, wmspec):
        topLevelTask = getFirstTask(wmspec)
        inputDataset = topLevelTask.inputDataset()
        dbs = DBSReader(inputDataset.dbsurl)
        # dbsDict = {self.inputDataset.dbsurl : self.dbs}
        return dbs

    def createWMBSHelperWithTopTask(self, wmspec, block, mask=None, parentFlag=False,
                                    detail=False, commonLocation=None):

        topLevelTask = getFirstTask(wmspec)

        wmbs = WMBSHelper(wmspec, topLevelTask.name(), block, mask, cachepath=self.workDir,
                          commonLocation=commonLocation)
        if block:
            blockName = block
            if parentFlag:
                block = self.dbs.getFileBlockWithParents(blockName)
                data = self.rucio.getReplicaInfoForBlocks(block=[blockName])
                block['PhEDExNodeNames'] = data[0]["replica"]
            else:
                block = self.dbs.getFileBlock(blockName)
                data = self.rucio.getReplicaInfoForBlocks(block=[blockName])
                block['PhEDExNodeNames'] = data[0]["replica"]
        sub, files = wmbs.createSubscriptionAndAddFiles(block=block)
        if detail:
            return wmbs, sub, files
        else:
            return wmbs

    def testKillWorkflow(self):
        """
        _testKillWorkflow_

        Verify that workflow killing works correctly.
        """
        baAPI = BossAirAPI(config=self.config, insertStates=True)

        # Create nine jobs
        self.setupForKillTest(baAPI=baAPI)
        self.assertEqual(len(baAPI._listRunJobs()), 9)
        killWorkflow("Main", self.config, self.config)

        self.verifyFileKillStatus()
        self.verifyJobKillStatus()
        self.assertEqual(len(baAPI._listRunJobs()), 8)

        return

    def testCreateSubscription(self):
        """
        _testCreateSubscription_

        Verify that the subscription creation code works correctly.
        """
        resourceControl = ResourceControl()
        resourceControl.insertSite(siteName='site1', pnn='goodse.cern.ch',
                                   ceName='site1', plugin="TestPlugin")
        resourceControl.insertSite(siteName='site2', pnn='goodse2.cern.ch',
                                   ceName='site2', plugin="TestPlugin")

        testWorkload = self.createTestWMSpec()
        testTopLevelTask = getFirstTask(testWorkload)
        testWMBSHelper = WMBSHelper(testWorkload, testTopLevelTask.name(), "SomeBlock", cachepath=self.workDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testTopLevelTask, testWMBSHelper.topLevelFileset)

        procWorkflow = Workflow(name="TestWorkload",
                                task="/TestWorkload/ProcessingTask")
        procWorkflow.load()

        self.assertEqual(procWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner: %s" % procWorkflow.owner)
        self.assertEqual(procWorkflow.group, "DMWM",
                         "Error: Wrong group: %s" % procWorkflow.group)
        self.assertEqual(procWorkflow.wfType, "TestReReco",
                         "Error: Wrong type.")
        self.assertEqual(procWorkflow.spec, os.path.join(self.workDir, procWorkflow.name,
                                                         "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(procWorkflow.outputMap), 1,
                         "Error: Wrong number of WF outputs.")
        mergedProcOutput = procWorkflow.outputMap["OutputADataTierA"][0]["merged_output_fileset"]
        unmergedProcOutput = procWorkflow.outputMap["OutputADataTierA"][0]["output_fileset"]

        mergedProcOutput.loadData()
        unmergedProcOutput.loadData()
        self.assertEqual(mergedProcOutput.name, "/TestWorkload/ProcessingTask/MergeTask/merged-MergedDataTierA",
                         "Error: Merged output fileset is wrong.")
        self.assertEqual(unmergedProcOutput.name, "/TestWorkload/ProcessingTask/unmerged-OutputADataTierA",
                         "Error: Unmerged output fileset is wrong.")

        mergeWorkflow = Workflow(name="TestWorkload",
                                 task="/TestWorkload/ProcessingTask/MergeTask")
        mergeWorkflow.load()

        self.assertEqual(mergeWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner.")
        self.assertEqual(mergeWorkflow.spec, os.path.join(self.workDir, mergeWorkflow.name,
                                                          "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(mergeWorkflow.outputMap), 1,
                         "Error: Wrong number of WF outputs.")

        cleanupWorkflow = Workflow(name="TestWorkload",
                                   task="/TestWorkload/ProcessingTask/CleanupTask")
        cleanupWorkflow.load()

        self.assertEqual(cleanupWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner.")
        self.assertEqual(cleanupWorkflow.spec, os.path.join(self.workDir, cleanupWorkflow.name,
                                                            "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(cleanupWorkflow.outputMap), 0,
                         "Error: Wrong number of WF outputs.")

        unmergedMergeOutput = mergeWorkflow.outputMap["MergedDataTierA"][0]["output_fileset"]
        unmergedMergeOutput.loadData()

        self.assertEqual(unmergedMergeOutput.name, "/TestWorkload/ProcessingTask/MergeTask/merged-MergedDataTierA",
                         "Error: Unmerged output fileset is wrong.")

        skimWorkflow = Workflow(name="TestWorkload",
                                task="/TestWorkload/ProcessingTask/MergeTask/SkimTask")
        skimWorkflow.load()

        self.assertEqual(skimWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner.")
        self.assertEqual(skimWorkflow.spec, os.path.join(self.workDir, skimWorkflow.name,
                                                         "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(skimWorkflow.outputMap), 2,
                         "Error: Wrong number of WF outputs.")

        mergedSkimOutputA = skimWorkflow.outputMap["SkimOutputADataTierA"][0]["merged_output_fileset"]
        unmergedSkimOutputA = skimWorkflow.outputMap["SkimOutputADataTierA"][0]["output_fileset"]
        mergedSkimOutputB = skimWorkflow.outputMap["SkimOutputBDataTierB"][0]["merged_output_fileset"]
        unmergedSkimOutputB = skimWorkflow.outputMap["SkimOutputBDataTierB"][0]["output_fileset"]

        mergedSkimOutputA.loadData()
        mergedSkimOutputB.loadData()
        unmergedSkimOutputA.loadData()
        unmergedSkimOutputB.loadData()

        self.assertEqual(mergedSkimOutputA.name,
                         "/TestWorkload/ProcessingTask/MergeTask/SkimTask/unmerged-SkimOutputADataTierA",
                         "Error: Merged output fileset is wrong: %s" % mergedSkimOutputA.name)
        self.assertEqual(unmergedSkimOutputA.name,
                         "/TestWorkload/ProcessingTask/MergeTask/SkimTask/unmerged-SkimOutputADataTierA",
                         "Error: Unmerged output fileset is wrong.")
        self.assertEqual(mergedSkimOutputB.name,
                         "/TestWorkload/ProcessingTask/MergeTask/SkimTask/unmerged-SkimOutputBDataTierB",
                         "Error: Merged output fileset is wrong.")
        self.assertEqual(unmergedSkimOutputB.name,
                         "/TestWorkload/ProcessingTask/MergeTask/SkimTask/unmerged-SkimOutputBDataTierB",
                         "Error: Unmerged output fileset is wrong.")

        topLevelFileset = Fileset(name="TestWorkload-ProcessingTask-SomeBlock")
        topLevelFileset.loadData()

        procSubscription = Subscription(fileset=topLevelFileset, workflow=procWorkflow)
        procSubscription.loadData()

        self.assertEqual(len(procSubscription.getWhiteBlackList()), 2,
                         "Error: Wrong site white/black list for proc sub.")
        for site in procSubscription.getWhiteBlackList():
            if site["site_name"] == "site1":
                self.assertEqual(site["valid"], 1,
                                 "Error: Site should be white listed.")
            else:
                self.assertEqual(site["valid"], 0,
                                 "Error: Site should be black listed.")

        self.assertEqual(procSubscription["type"], "Processing",
                         "Error: Wrong subscription type.")
        self.assertEqual(procSubscription["split_algo"], "FileBased",
                         "Error: Wrong split algo.")

        mergeSubscription = Subscription(fileset=unmergedProcOutput, workflow=mergeWorkflow)
        mergeSubscription.loadData()

        self.assertEqual(len(mergeSubscription.getWhiteBlackList()), 0,
                         "Error: Wrong white/black list for merge sub.")

        self.assertEqual(mergeSubscription["type"], "Merge",
                         "Error: Wrong subscription type.")
        self.assertEqual(mergeSubscription["split_algo"], "WMBSMergeBySize",
                         "Error: Wrong split algo.")

        skimSubscription = Subscription(fileset=unmergedMergeOutput, workflow=skimWorkflow)
        skimSubscription.loadData()

        self.assertEqual(skimSubscription["type"], "Skim",
                         "Error: Wrong subscription type.")
        self.assertEqual(skimSubscription["split_algo"], "FileBased",
                         "Error: Wrong split algo.")
        return

    def testTruncatedWFInsertion(self):
        """
        _testTruncatedWFInsertion_

        """
        resourceControl = ResourceControl()
        resourceControl.insertSite(siteName='site1', pnn='goodse.cern.ch',
                                   ceName='site1', plugin="TestPlugin")
        resourceControl.insertSite(siteName='site2', pnn='goodse2.cern.ch',
                                   ceName='site2', plugin="TestPlugin")

        testWorkload = self.createTestWMSpec()
        testTopLevelTask = getFirstTask(testWorkload)
        testWMBSHelper = WMBSHelper(testWorkload, testTopLevelTask.name(), "SomeBlock", cachepath=self.workDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testTopLevelTask, testWMBSHelper.topLevelFileset)

        testWorkload.truncate("ResubmitTestWorkload", "/TestWorkload/ProcessingTask/MergeTask",
                              "someserver", "somedatabase")

        # create  the subscription for multiple top task (MergeTask and CleanupTask for the same block)
        for task in testWorkload.getTopLevelTask():
            testResubmitWMBSHelper = WMBSHelper(testWorkload, task.name(), "SomeBlock2", cachepath=self.workDir)
            testResubmitWMBSHelper.createTopLevelFileset()
            testResubmitWMBSHelper._createSubscriptionsInWMBS(task, testResubmitWMBSHelper.topLevelFileset)

        mergeWorkflow = Workflow(name="ResubmitTestWorkload",
                                 task="/ResubmitTestWorkload/MergeTask")
        mergeWorkflow.load()

        self.assertEqual(mergeWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner.")
        self.assertEqual(mergeWorkflow.spec, os.path.join(self.workDir, mergeWorkflow.name,
                                                          "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(mergeWorkflow.outputMap), 1, "Error: Wrong number of WF outputs.")

        unmergedMergeOutput = mergeWorkflow.outputMap["MergedDataTierA"][0]["output_fileset"]
        unmergedMergeOutput.loadData()

        self.assertEqual(unmergedMergeOutput.name, "/ResubmitTestWorkload/MergeTask/merged-MergedDataTierA",
                         "Error: Unmerged output fileset is wrong.")

        skimWorkflow = Workflow(name="ResubmitTestWorkload",
                                task="/ResubmitTestWorkload/MergeTask/SkimTask")
        skimWorkflow.load()

        self.assertEqual(skimWorkflow.owner, "sfoulkes",
                         "Error: Wrong owner.")
        self.assertEqual(skimWorkflow.spec, os.path.join(self.workDir, skimWorkflow.name,
                                                         "WMSandbox", "WMWorkload.pkl"),
                         "Error: Wrong spec URL")
        self.assertEqual(len(skimWorkflow.outputMap), 2,
                         "Error: Wrong number of WF outputs.")

        mergedSkimOutputA = skimWorkflow.outputMap["SkimOutputADataTierA"][0]["merged_output_fileset"]
        unmergedSkimOutputA = skimWorkflow.outputMap["SkimOutputADataTierA"][0]["output_fileset"]
        mergedSkimOutputB = skimWorkflow.outputMap["SkimOutputBDataTierB"][0]["merged_output_fileset"]
        unmergedSkimOutputB = skimWorkflow.outputMap["SkimOutputBDataTierB"][0]["output_fileset"]

        mergedSkimOutputA.loadData()
        mergedSkimOutputB.loadData()
        unmergedSkimOutputA.loadData()
        unmergedSkimOutputB.loadData()

        self.assertEqual(mergedSkimOutputA.name,
                         "/ResubmitTestWorkload/MergeTask/SkimTask/unmerged-SkimOutputADataTierA",
                         "Error: Merged output fileset is wrong: %s" % mergedSkimOutputA.name)
        self.assertEqual(unmergedSkimOutputA.name,
                         "/ResubmitTestWorkload/MergeTask/SkimTask/unmerged-SkimOutputADataTierA",
                         "Error: Unmerged output fileset is wrong.")
        self.assertEqual(mergedSkimOutputB.name,
                         "/ResubmitTestWorkload/MergeTask/SkimTask/unmerged-SkimOutputBDataTierB",
                         "Error: Merged output fileset is wrong.")
        self.assertEqual(unmergedSkimOutputB.name,
                         "/ResubmitTestWorkload/MergeTask/SkimTask/unmerged-SkimOutputBDataTierB",
                         "Error: Unmerged output fileset is wrong.")

        topLevelFileset = Fileset(name="ResubmitTestWorkload-MergeTask-SomeBlock2")
        topLevelFileset.loadData()

        mergeSubscription = Subscription(fileset=topLevelFileset, workflow=mergeWorkflow)
        mergeSubscription.loadData()

        self.assertEqual(len(mergeSubscription.getWhiteBlackList()), 0,
                         "Error: Wrong white/black list for merge sub.")

        self.assertEqual(mergeSubscription["type"], "Merge",
                         "Error: Wrong subscription type.")
        self.assertEqual(mergeSubscription["split_algo"], "WMBSMergeBySize",
                         "Error: Wrong split algo.")

        skimSubscription = Subscription(fileset=unmergedMergeOutput, workflow=skimWorkflow)
        skimSubscription.loadData()

        self.assertEqual(skimSubscription["type"], "Skim",
                         "Error: Wrong subscription type.")
        self.assertEqual(skimSubscription["split_algo"], "FileBased",
                         "Error: Wrong split algo.")

        return

    def testReReco(self):
        """ReReco workflow"""
        # create workflow
        block = self.dataset + "#" + BLOCK1
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        files = wmbs.validFiles(self.dbs.getFileBlock(block)['Files'])
        self.assertEqual(len(files), 5)

    def testReRecoBlackRunRestriction(self):
        """ReReco workflow with Run restrictions"""
        block = self.dataset + "#" + BLOCK2
        self.topLevelTask.setInputRunBlacklist([181183])  # Set run blacklist to only run in the block
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)

        files = wmbs.validFiles(self.dbs.getFileBlock(block)['Files'])
        self.assertEqual(len(files), 0)

    def testReRecoWhiteRunRestriction(self):
        block = self.dataset + "#" + BLOCK2
        self.topLevelTask.setInputRunWhitelist([181183])  # Set run whitelist to only run in the block
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        files = wmbs.validFiles(self.dbs.getFileBlock(block)['Files'])
        self.assertEqual(len(files), 1)

    def testLumiMaskRestrictionsOK(self):
        block = self.dataset + "#" + BLOCK1
        self.wmspec.getTopLevelTask()[0].data.input.splitting.runs = ['181367']
        self.wmspec.getTopLevelTask()[0].data.input.splitting.lumis = ['57,80']
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        files = wmbs.validFiles(self.dbs.getFileBlock(block)['Files'])
        self.assertEqual(len(files), 1)

    def testLumiMaskRestrictionsKO(self):
        block = self.dataset + "#" + BLOCK1
        self.wmspec.getTopLevelTask()[0].data.input.splitting.runs = ['123454321']
        self.wmspec.getTopLevelTask()[0].data.input.splitting.lumis = ['123,123']
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        files = wmbs.validFiles(self.dbs.getFileBlock(block)['Files'])
        self.assertEqual(len(files), 0)

    def testDuplicateFileInsert(self):
        # using default wmspec
        block = self.dataset + "#" + BLOCK1
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        wmbs.topLevelFileset.loadData()
        numOfFiles = len(wmbs.topLevelFileset.files)
        # check initially inserted files.
        dbsFiles = self.dbs.getFileBlock(block)['Files']
        self.assertEqual(numOfFiles, len(dbsFiles))
        firstFileset = wmbs.topLevelFileset
        wmbsDao = wmbs.daofactory(classname="Files.InFileset")

        numOfFiles = len(wmbsDao.execute(firstFileset.id))
        self.assertEqual(numOfFiles, len(dbsFiles))

        # use the new spec with same inputdataset
        block = self.dataset + "#" + BLOCK1
        wmspec = self.createWMSpec("TestSpec1")
        dbs = self.getDBS(wmspec)
        wmbs = self.createWMBSHelperWithTopTask(wmspec, block)
        # check duplicate insert
        dbsFiles = dbs.getFileBlock(block)
        data = self.rucio.getReplicaInfoForBlocks(block=[block])
        dbsFiles['PhEDExNodeNames'] = data[0]["replica"]
        numOfFiles = wmbs.addFiles(dbsFiles)
        self.assertEqual(numOfFiles, 0)
        secondFileset = wmbs.topLevelFileset

        wmbsDao = wmbs.daofactory(classname="Files.InFileset")
        numOfFiles = len(wmbsDao.execute(secondFileset.id))
        self.assertEqual(numOfFiles, len(dbsFiles['Files']))

        self.assertNotEqual(firstFileset.id, secondFileset.id)

    def testDuplicateSubscription(self):
        """Can't duplicate subscriptions"""
        siteWhitelist = ["T2_XX_SiteA", "T2_XX_SiteB"]
        # using default wmspec
        block = self.dataset + "#" + BLOCK1
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        wmbs.topLevelFileset.loadData()
        numOfFiles = len(wmbs.topLevelFileset.files)
        filesetId = wmbs.topLevelFileset.id
        subId = wmbs.topLevelSubscription['id']

        # check initially inserted files.
        dbsFiles = self.dbs.getFileBlock(block)['Files']
        self.assertEqual(numOfFiles, len(dbsFiles))

        # Not clear what's supposed to happen here, 2nd test is completely redundant
        dummyFirstFileset = wmbs.topLevelFileset
        self.assertEqual(numOfFiles, len(dbsFiles))

        # reinsert subscription - shouldn't create anything new
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, block)
        wmbs.topLevelFileset.loadData()
        self.assertEqual(numOfFiles, len(wmbs.topLevelFileset.files))
        self.assertEqual(filesetId, wmbs.topLevelFileset.id)
        self.assertEqual(subId, wmbs.topLevelSubscription['id'])

        # now do a montecarlo workflow
        self.setupMCWMSpec()
        mask = Mask(FirstRun=12, FirstLumi=1234, FirstEvent=12345,
                    LastEvent=999995, LastLumi=12345, LastRun=12)
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, None, mask,
                                                commonLocation=siteWhitelist)
        wmbs.topLevelFileset.loadData()
        numOfFiles = len(wmbs.topLevelFileset.files)
        filesetId = wmbs.topLevelFileset.id
        subId = wmbs.topLevelSubscription['id']

        # check initially inserted files.
        # Not clear what's supposed to happen here, 2nd test is completely redundant
        numDbsFiles = 1
        self.assertEqual(numOfFiles, numDbsFiles)
        dummyFirstFileset = wmbs.topLevelFileset
        self.assertEqual(numOfFiles, numDbsFiles)

        # reinsert subscription - shouldn't create anything new
        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, None, mask,
                                                commonLocation=siteWhitelist)
        wmbs.topLevelFileset.loadData()
        self.assertEqual(numOfFiles, len(wmbs.topLevelFileset.files))
        self.assertEqual(filesetId, wmbs.topLevelFileset.id)
        self.assertEqual(subId, wmbs.topLevelSubscription['id'])

    def testParentage(self):
        """
        1. check whether parent files are created in wmbs.
        2. check parent files are associated to child.
        3. When 2 specs with the same input data (one with parent processing, one without it)
           is inserted, if one without parent processing inserted first then the other with
           parent processing insert, it still needs to create parent files although child files
           are duplicate
        """

        # Swap out the dataset for one that has parents
        task = next(self.wmspec.taskIterator())
        oldDS = task.inputDataset()  # Copy the old dataset, only will use DBS URL from it
        task.addInputDataset(name="/Cosmics/ComissioningHI-PromptReco-v1/RECO",
                             primary='Cosmics', processed='ComissioningHI-PromptReco-v1',
                             tier='RECO', dbsurl=oldDS.dbsurl)
        block = '/Cosmics/ComissioningHI-PromptReco-v1/RECO' + '#5b89ba9c-0dbf-11e1-9b6c-003048caaace'

        # File creation without parents
        wmbs, _, numFiles = self.createWMBSHelperWithTopTask(self.wmspec, block, parentFlag=False, detail=True)
        self.assertEqual(8, numFiles)
        wmbs.topLevelFileset.loadData()
        for child in wmbs.topLevelFileset.files:
            self.assertEqual(len(child["parents"]), 0)  # no parents per child

        # File creation with parents
        wmbs, _, numFiles = self.createWMBSHelperWithTopTask(self.wmspec, block, parentFlag=True, detail=True)
        self.assertEqual(8, numFiles)
        wmbs.topLevelFileset.loadData()
        for child in wmbs.topLevelFileset.files:
            self.assertEqual(len(child["parents"]), 1)  # one parent per child

    def testMCFakeFileInjection(self):
        """Inject fake Monte Carlo files into WMBS"""

        # This test is failing because the name of the couch DB is set to None
        # in BasicProductionWorkload.getProdArgs() but changing it to
        # "reqmgr_config_cache_t" from StdBase test arguments does not fix the
        # situation. testDuplicateSubscription probably has the same issue
        siteWhitelist = ["T2_XX_SiteA", "T2_XX_SiteB"]

        self.setupMCWMSpec()

        mask = Mask(FirstRun=12, FirstLumi=1234, FirstEvent=12345,
                    LastEvent=999995, LastLumi=12345, LastRun=12)

        wmbs = self.createWMBSHelperWithTopTask(self.wmspec, None, mask,
                                                commonLocation=siteWhitelist)
        subscription = wmbs.topLevelSubscription
        self.assertEqual(1, subscription.exists())
        fileset = subscription['fileset']
        self.assertEqual(1, fileset.exists())
        fileset.loadData()  # need to refresh from database

        self.assertEqual(len(fileset.files), 1)
        self.assertEqual(len(fileset.parents), 0)
        self.assertFalse(fileset.open)

        firstFile = list(fileset.files)[0]
        self.assertEqual(firstFile['events'], mask['LastEvent'] - mask['FirstEvent'] + 1)  # inclusive range
        self.assertEqual(firstFile['merged'], False)  # merged files get added to dbs
        self.assertEqual(len(firstFile['parents']), 0)
        # firstFile.loadData()
        self.assertEqual(sorted(firstFile['locations']), sorted(self.pnns))
        self.assertEqual(len(firstFile.getParentLFNs()), 0)

        self.assertEqual(len(firstFile.getRuns()), 1)
        run = firstFile.getRuns()[0]
        self.assertEqual(run.run, mask['FirstRun'])
        self.assertEqual(run.lumis[0], mask['FirstLumi'])
        self.assertEqual(run.lumis[-1], mask['LastLumi'])
        self.assertEqual(len(run.lumis), mask['LastLumi'] - mask['FirstLumi'] + 1)


if __name__ == '__main__':
    unittest.main()



#!/bin/env python
#pylint: disable=E1103, C0103
# E1103: Threads have things attached to them
# C0103: We use readable names for method names
"""
Unittest to see if the main configs need updating

"""

import re
import os.path
import logging
import unittest
import threading

from nose.plugins.attrib import attr

import WMCore.WMInit
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit

from WMCore.Agent.Configuration import loadConfigurationFile



# Now comes the ugly work
from WMComponent.JobCreator.JobCreator          import JobCreator
from WMComponent.JobSubmitter.JobSubmitter      import JobSubmitter
from WMComponent.JobAccountant.JobAccountant    import JobAccountant
from WMComponent.JobTracker.JobTracker          import JobTracker
from WMComponent.JobArchiver.JobArchiver        import JobArchiver
from WMComponent.TaskArchiver.TaskArchiver      import TaskArchiver
from WMComponent.ErrorHandler.ErrorHandler      import ErrorHandler
from WMComponent.RetryManager.RetryManager      import RetryManager
from WMComponent.DBS3Buffer.DBS3Upload          import DBS3Upload


class ConfigTest(unittest.TestCase):
    """
    Test class for the WMAgent configs

    """


    def setUp(self):
        """
        _setUp_

        Create the whole database
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase = True)
        self.testInit.setSchema(customModules = ["WMCore.WMBS",
                                                 "WMComponent.DBS3Buffer",
                                                 'WMCore.ResourceControl',
                                                 'WMCore.BossAir'],
                                useDefault = False)
        self.dbName = 'config_t'
        self.testInit.setupCouch("%s/jobs" % self.dbName, "JobDump")
        self.testInit.setupCouch("%s/fwjrs" % self.dbName, "FWJRDump")
        self.testInit.setupCouch("%s/acdc" % self.dbName, "GroupUser", "ACDC")
        self.testInit.setupCouch("%s/workqueue" % self.dbName, 'WorkQueue')
        self.testInit.setupCouch("%s/workloadsummary" % self.dbName,
                                 "WorkloadSummary")

        self.testDir = self.testInit.generateWorkDir()
        return

    def tearDown(self):
        """
        _tearDown_

        Tear things down and go home
        """
        self.testInit.clearDatabase()
        self.testInit.tearDownCouch()
        return


    def initComponents(self, configPath):
        """
        _initComponents_

        Start up the various components using the config
        """

        if os.path.isfile(configPath):
            # Read the config
            config = loadConfigurationFile(configPath)
        else:
            msg = "No config file at desired location"
            logging.error(msg)
            raise Exception(msg)


        masterConfig = self.testInit.getConfiguration()
        config.Agent.useHeartbeat = False
        config.CoreDatabase.socket     = masterConfig.CoreDatabase.socket
        config.CoreDatabase.connectUrl = masterConfig.CoreDatabase.connectUrl


        # Have to do this because the agent hard codes its dirs
        oldWorkDir = config.General.workDir

        for compName in (config.listComponents_() + config.listWebapps_()):
            component = getattr(config, compName)
            for var in component.listSections_():
                value = getattr(component, var)
                if type(value) == str:
                    if re.search(oldWorkDir, value):
                        # Then set it
                        setattr(component, var, value.replace(oldWorkDir, self.testDir))
                elif type(value) == list:
                    # Go through it one component at a time
                    for element in value:
                        if type(element) == str and re.search(oldWorkDir, element):
                            index = value.index(element)
                            value.remove(element)
                            value.insert(index, value.replace(oldWorkDir, self.testDir))
                    setattr(component, var, value)
                elif type(value) == dict:
                    for key in value.keys():
                        if type(value[key]) == str and re.search(oldWorkDir, value[key]):
                            value[key] = value[key].replace(oldWorkDir, self.testDir)
                    setattr(component, var, value)


        compList = (config.listComponents_() + config.listWebapps_())
        components = []

        config.JobStateMachine.couchurl                = os.environ['COUCHURL']
        config.JobStateMachine.couchDBName             = self.dbName
        config.ACDC.couchurl                           = os.environ['COUCHURL']
        config.ACDC.database                           = '%s/acdc' % self.dbName
        config.TaskArchiver.workloadSummaryCouchDBName = '%s/workloadsummary' % self.dbName
        config.TaskArchiver.workloadSummaryCouchURL    = os.environ['COUCHURL']

        if hasattr(config, 'WorkQueueManager'):
            config.WorkQueueManager.couchurl   = os.environ['COUCHURL']
            config.WorkQueueManager.dbname     = '%s/workqueue' % self.dbName

        if hasattr(config, 'WorkloadSummary'):
            config.WorkloadSummary.couchurl    = os.environ['COUCHURL']
            config.WorkloadSummary.database    = '%s/workloadsummary' % self.dbName


        # Get all components


        components.append(JobCreator(config = config))
        components.append(JobSubmitter(config = config))
        components.append(JobTracker(config = config))
        components.append(JobAccountant(config = config))
        components.append(JobArchiver(config = config))
        components.append(TaskArchiver(config = config))
        components.append(ErrorHandler(config = config))
        components.append(RetryManager(config = config))
        components.append(DBS3Upload(config = config))

        # Init threads:
        for component in components:
            component.initInThread()

        # preInitialize
        for component in components:
            component.preInitialization()

        for component in components:
            component.prepareToStop()

        return


    @attr('integration')
    def testA_WMAgentConfig(self):
        """
        _WMAgentConfig_

        Test the WMAgentConfig file in WMCORE/etc
        """

        # Get the config
        configPath = os.path.join(WMCore.WMInit.getWMBASE(),
                                  'etc', 'WMAgentConfig.py')


        self.initComponents(configPath = configPath)


        return

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
    WorkQueue.Policy.End tests
"""




import unittest
from functools import partial
from WMCore.WorkQueue.Policy.End.EndPolicyInterface import EndPolicyInterface
from WMCore.WorkQueue.DataStructs.WorkQueueElement import WorkQueueElement as WQE


class EndPolicyInterfaceTestCase(unittest.TestCase):

    def setUp(self):
        """Create workflow stuff"""
        self.policy = partial(EndPolicyInterface)

        # ones i made earlier
        self.parent = WQE(); self.parent.id = 1
        self.available = WQE(Status = 'Available', ParentQueueId = 1)
        self.acquired = WQE(Status = 'Acquired', ParentQueueId = 1)
        self.negotiating = WQE(Status = 'Negotiating', ParentQueueId = 1)
        self.done = WQE(Status = 'Done', PercentComplete = 100, PercentSuccess = 100, ParentQueueId = 1)
        self.failed = WQE(Status = 'Failed', PercentComplete = 100, PercentSuccess = 0, ParentQueueId = 1)

    def tearDown(self):
        pass

    def testOverallStatus(self):
        """Overall status correct for given inputs"""

        # check no final state till all elements in end state
        for items in ((self.available, self.available),
                      (self.acquired, self.available),
                      (self.available, self.acquired),
                      (self.available, self.failed),
                      (self.available, self.done),
                      (self.available, self.negotiating),
                      ):
            results = self.policy()(items, [self.parent])
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['Status'], 'Acquired')

    def testEndConditions(self):
        """Correct status when all elements in an end state"""
        for result in (('Done', self.done, self.done),
                       ('Failed', self.failed, self.done)):
            self.assertEqual(result[0], self.policy()(result[1:], [self.parent])[0]['Status'])

    def testNegotiating(self):
        """Workflow only partially split"""
        # one element split and available in workqueue, other not split yet
        parent2 = WQE(Status = 'Negotiating')
        parent2.id = 2
        parents = [self.parent, parent2]
        results = self.policy()([self.available], parents)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]['Status'], 'Acquired')
        self.assertEqual(results[1]['Status'], 'Negotiating')

if __name__ == '__main__':
    unittest.main()



"""
Hello world example using WMCore.REST handling framework.
Info class giving information about ReqMgr database.
Teams, Groups, Software versions handling for ReqMgr.

"""
from __future__ import print_function, division
from future.utils import listvalues

import json
import cherrypy

import WMCore
from WMCore.Database.CMSCouch import Document, CouchNotFoundError, CouchError
from WMCore.REST.Error import RESTError, NoSuchInstance, APINotSpecified
from WMCore.REST.Format import JSONFormat, PrettyJSONFormat
from WMCore.REST.Server import RESTEntity, restcall, rows
from WMCore.REST.Tools import tools
from WMCore.ReqMgr.DataStructs.ReqMgrConfigDataCache import ReqMgrConfigDataCache
from WMCore.Services.RequestDB.RequestDBReader import RequestDBReader


class MissingPostData(RESTError):
    "The specified object is invalid."
    http_code = 400
    app_code = 1201

    def __init__(self):
        RESTError.__init__(self)
        self.message = "Empty data has passed"


class Info(RESTEntity):
    """
    general information about reqmgr2, i.e. version, etc
    """

    def __init__(self, app, api, config, mount):
        RESTEntity.__init__(self, app, api, config, mount)
        self.reqmgr_db = api.db_handler.get_db(config.couch_reqmgr_db)
        self.config = config

    def validate(self, apiobj, method, api, param, safe):
        pass

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        # authorization / access control:
        # cherrypy (modified CMS web version here) can read this information
        # from HTTP headers. CMS web frontend puts this information into
        # headers as read from CRIC
        #        print "cherrypy.request", cherrypy.request
        #        print "DN: %s" % cherrypy.request.user['dn']
        #        print "Requestor/login: %s" % cherrypy.request.user['login']
        #        print "cherrypy.request: %s" % cherrypy.request
        #        print "cherrypy.request.user: %s" % cherrypy.request.user
        # from WMCore.REST.Auth import authz_match
        # authz_match(role=["Global Admin"], group=["global"])

        # implement as authentication decorator over modification calls
        # check config.py main.authz_defaults and CRIC
        # (only Admin: ReqMgr to be able to modify stuff)

        wmcore_reqmgr_version = WMCore.__version__

        reqmgr_db_info = self.reqmgr_db.info()
        reqmgr_db_info["reqmgr_couch_url"] = self.config.couch_host

        # retrieve the last injected request in the system
        # curl ... /reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1
        options = {"descending": True, "limit": 1}
        reqmgr_last_injected_request = self.reqmgr_db.loadView("ReqMgr", "bydate", options=options)
        result = {"wmcore_reqmgr_version": wmcore_reqmgr_version,
                  "reqmgr_db_info": reqmgr_db_info,
                  "reqmgr_last_injected_request": reqmgr_last_injected_request}
        # NOTE:
        # "return result" only would return dict with keys without (!!) values set
        return rows([result])


class ReqMgrConfigData(RESTEntity):
    def __init__(self, app, api, config, mount):
        # CouchDB auxiliary database name
        RESTEntity.__init__(self, app, api, config, mount)
        self.reqmgr_aux_db = api.db_handler.get_db(config.couch_reqmgr_aux_db)

    def _validate_args(self, param, safe):
        # TODO: need proper validation but for now pass everything
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["doc_name"] = param.args[0]
            param.args.pop()
        else:
            raise APINotSpecified()

        return

    def _validate_put_args(self, param, safe):

        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["doc_name"] = param.args[0]
            param.args.pop()

        data = cherrypy.request.body.read()

        if data:
            config_args = json.loads(data)
            # TODO need to validate the args depending on the config
            safe.kwargs["config_dict"] = config_args

    def validate(self, apiobj, method, api, param, safe):
        """
        Validate request input data.
        Has to be implemented, otherwise the service fails to start.
        If it's not implemented correctly (e.g. just pass), the arguments
        are not passed in the method at all.

        """
        if method == "GET":
            self._validate_args(param, safe)
        elif method == "PUT":
            self._validate_put_args(param, safe)

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    def get(self, doc_name):
        """
        """
        config = ReqMgrConfigDataCache.getConfig(doc_name)
        return rows([config])

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    def put(self, doc_name, config_dict=None):
        """
        """
        if doc_name == "DEFAULT":
            return ReqMgrConfigDataCache.putDefaultConfig()
        return ReqMgrConfigDataCache.replaceConfig(doc_name, config_dict)


class AuxBaseAPI(RESTEntity):
    """
    Base class for Aux db RESTEntry which contains get, post method
    """

    def __init__(self, app, api, config, mount):
        RESTEntity.__init__(self, app, api, config, mount)
        # CouchDB auxiliary database name
        self.reqmgr_aux_db = api.db_handler.get_db(config.couch_reqmgr_aux_db)
        self.reqmgr_aux_db_service = RequestDBReader(self.reqmgr_aux_db, couchapp="ReqMgrAux")
        self.setName()

    def setName(self):
        "Sets the document name"
        raise NotImplementedError("Couch document id(name) should be specified. i.e. self.name='software'")

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["subName"] = param.args.pop(0)
            return
        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    def get(self, subName=None):
        """
        Return entire self.name document
        subName is subcategory of document which is added as postfix string
        """
        try:
            if subName:
                if subName.lower() == "all_docs":
                    return rows(self._getAllDocs())
                else:
                    docName = "%s_%s" % (self.name, subName)
            else:
                docName = self.name
            sw = self.reqmgr_aux_db.document(docName)
            del sw["_id"]
            del sw["_rev"]
        except CouchNotFoundError:
            raise NoSuchInstance

        return rows([sw])

    def _getAllDocs(self):
        """
        Return all the documents under a given document type
        """
        try:
            option = {"include_docs": True}
            allDocs = self.reqmgr_aux_db_service.getRequestByCouchView("byconfig", option, [self.name])
        except CouchError as ex:
            msg = "ERROR: Failed to fetch ALL_DOCS for ConfigType: %s. Reason: %s" % (self.name, str(ex))
            cherrypy.log(msg)
            raise cherrypy.HTTPError(404, msg)

        return listvalues(allDocs)

    @restcall(formats=[('application/json', JSONFormat())])
    def post(self, subName=None):
        """
        Inserts a new document into the database
        """
        data = cherrypy.request.body.read()
        if not data:
            raise MissingPostData()
        else:
            doc = json.loads(data)
        if subName:
            docName = "%s_%s" % (self.name, subName)
        else:
            docName = self.name

        doc["ConfigType"] = self.name
        doc = Document(docName, doc)
        result = self.reqmgr_aux_db.commitOne(doc)
        return result

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    def put(self, subName=None):
        """
        Update document for the given self.name and subName.
        It assumes the client has provided the entire entity, i.e., the old
        content gets completely replaced by the new one.

        Given that the each couch document contains a revision number, these PUT calls
        are not going to be idempotent.
        """
        data = cherrypy.request.body.read()
        if not data:
            raise MissingPostData()
        else:
            propertyDict = json.loads(data)

        result = None
        if subName:
            docName = "%s_%s" % (self.name, subName)
        else:
            docName = self.name

        try:
            existDoc = self.reqmgr_aux_db.document(docName)
            # replace original document
            newDoc = Document(existDoc['_id'], inputDict={'_rev': existDoc['_rev'],
                                                          'ConfigType': existDoc['ConfigType']})
            newDoc.update(propertyDict)
            result = self.reqmgr_aux_db.commitOne(newDoc)
        except CouchNotFoundError:
            cherrypy.log("Document %s not found. Creating one." % docName)
            doc = Document(docName, propertyDict)
            doc.update({'ConfigType': self.name})
            result = self.reqmgr_aux_db.commitOne(doc)

        return result

    @restcall(formats=[('application/json', JSONFormat())])
    def delete(self, subName):
        """
        Delete a document from ReqMgrAux
        """
        docName = "%s_%s" % (self.name, subName)
        try:
            res = self.reqmgr_aux_db.delete_doc(docName)
        except (CouchError, CouchNotFoundError) as ex:
            msg = "ERROR: failed to delete document: %s\nReason: %s" % (docName, str(ex))
            cherrypy.log(msg)
            res = None
        return res


class CMSSWVersions(AuxBaseAPI):
    """
    Handle CMSSW versions and scram architectures.
    Stored in the ReqMgr reqmgr_auxiliary database,
    """

    def setName(self):
        self.name = "CMSSW_VERSIONS"


class WMAgentConfig(AuxBaseAPI):
    """
    Handle WMAgent configuration files, which is
    a sub-set of the common config.py agent configuration
    """

    def setName(self):
        self.name = "WMAGENT_CONFIG"


class CampaignConfig(AuxBaseAPI):
    """
    Handle central production campaign configuration documents.
    """

    def setName(self):
        self.name = "CAMPAIGN_CONFIG"


class UnifiedConfig(AuxBaseAPI):
    """
    Handle central production unified configuration
    """

    def setName(self):
        self.name = "UNIFIED_CONFIG"


class TransferInfo(AuxBaseAPI):
    """
    Handle information related to the automatic transfers made
    by the MicroServices.
    """

    def setName(self):
        self.name = "TRANSFER"


class ParentLocks(AuxBaseAPI):
    """
    Handle list of locked parent datasets
    """

    def setName(self):
        self.name = "PARENT_LOCKS"



"""
Unit test for pycurl_manager module.
"""

from __future__ import division

import gzip
import os
import tempfile
import unittest
import traceback
from Utils.CertTools import getKeyCertFromEnv
from WMCore.Services.pycurl_manager import \
        RequestHandler, ResponseHeader, getdata, cern_sso_cookie, decompress


class PyCurlManager(unittest.TestCase):
    """Test pycurl_manager module"""

    def setUp(self):
        "initialization"
        self.mgr = RequestHandler()
        #self.ckey = os.path.join(os.environ['HOME'], '.globus/userkey.pem')
        #self.cert = os.path.join(os.environ['HOME'], '.globus/usercert.pem')
        self.ckey = getKeyCertFromEnv()[0]
        self.cert = getKeyCertFromEnv()[1]

        self.cricheader = 'Date: Tue, 06 Nov 2018 14:50:29 GMT\r\nServer: Apache/2.4.6 (CentOS) OpenSSL/1.0.2k-fips mod_wsgi/3.4 Python/2.7.5 mod_gridsite/2.3.4\r\nVary: Cookie\r\nX-Frame-Options: SAMEORIGIN\r\nSet-Cookie: sessionid=bc1xu8zi5rbbsd5fgjuklb2tk2r3f6tw; expires=Sun, 11-Nov-2018 14:50:29 GMT; httponly; Max-Age=432000; Path=/\r\nContent-Length: 32631\r\nContent-Type: application/json\r\n\r\n'
        self.dbsheader = 'Date: Tue, 06 Nov 2018 14:39:07 GMT\r\nServer: Apache\r\nCMS-Server-Time: D=1503 t=1541515147806112\r\nTransfer-Encoding: chunked\r\nContent-Type: text/html\r\n\r\n'
        self.HTTPheader = 'Date: Tue, 06 Nov 2018 14:50:29 GMT\r\nServer: Apache/2.4.6 (CentOS) OpenSSL/1.0.2k-fips mod_wsgi/3.4 Python/2.7.5 mod_gridsite/2.3.4\r\nVary: Cookie\r\nX-Frame-Options: SAMEORIGIN\r\nSet-Cookie: GRIDHTTP_PASSCODE=2c6da9c96efa2ad0farhda; domain=cms-cric.cern.ch; path=/; secure\r\nContent-Length: 32631\r\nContent-Type: application/json\r\n\r\n'

    def testDecompress(self):
        """
        Test decompress function
        """
        body = "bla"
        headers = {}
        data = decompress(body, headers)
        self.assertEqual(data, body)

        # gzip body
        gzipBody = gzip.compress(bytes(body, 'utf-8'))
        headers = {'Content-Encoding': 'gzip'}
        data = decompress(gzipBody, headers)
        self.assertEqual(data, bytes(body, 'utf-8'))

    def testMulti(self):
        """
        Test fetch of several urls at once, one of the url relies on CERN SSO.
        """
        tfile = tempfile.NamedTemporaryFile()
        url1 = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader/help"
        url2 = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader/datatiers"
        url3 = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
        cern_sso_cookie(url3, tfile.name, self.cert, self.ckey)
        cookie = {url3: tfile.name}
        urls = [url1, url2, url3]
        data = getdata(urls, self.ckey, self.cert, cookie=cookie)
        headers = 0
        for row in data:
            if '200 OK' in row['headers']:
                headers += 1
        self.assertTrue(headers, 3)

    def testSingle(self):
        """
        Test single call to CERN SSO url.
        """
        # test RequestHandler
        url = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
        params = {}
        headers = {"Cache-Control": "no-cache"}
        tfile = tempfile.NamedTemporaryFile()
        cern_sso_cookie(url, tfile.name, self.cert, self.ckey)
        cookie = {url: tfile.name}
        header, _ = self.mgr.request(url, params, headers, cookie=cookie)
        self.assertTrue(header.status, 200)

    def testContinue(self):
        """
        Test HTTP exit code 100 - Continue
        """
        header = "HTTP/1.1 100 Continue\r\n" + self.dbsheader

        resp = ResponseHeader(header)
        self.assertIsNone(getattr(resp, "status", None))
        self.assertEqual(resp.reason, "")
        self.assertFalse(resp.fromcache)
        self.assertIn("CMS-Server-Time", resp.header)
        self.assertIn("Date", resp.header)
        self.assertEqual(resp.header['Content-Type'], 'text/html')
        self.assertEqual(resp.header['Server'], 'Apache')
        self.assertEqual(resp.header['Transfer-Encoding'], 'chunked')
        return

    def testOK(self):
        """
        Test HTTP exit code 200 - OK
        """
        header = "HTTP/1.1 200 OK\r\n" + self.dbsheader

        resp = ResponseHeader(header)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.reason, "OK")
        self.assertFalse(resp.fromcache)
        return

    def testForbidden(self):
        """
        Test HTTP exit code 403 - Forbidden
        """
        header = "HTTP/1.1 403 Forbidden\r\n" + self.dbsheader

        resp = ResponseHeader(header)
        self.assertEqual(resp.status, 403)
        self.assertEqual(resp.reason, "Forbidden")
        self.assertFalse(resp.fromcache)
        return

    def testOKCRIC(self):
        """
        Test HTTP exit code 200 - OK for a CRIC response header
        """
        header = "HTTP/1.1 200 OK\r\n" + self.cricheader

        resp = ResponseHeader(header)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.reason, "OK")
        self.assertFalse(resp.fromcache)
        self.assertIn("Content-Length", resp.header)
        self.assertIn("Date", resp.header)
        self.assertIn("Server", resp.header)
        self.assertIn("sessionid", resp.header['Set-Cookie'])
        self.assertEqual(resp.header['Content-Type'], 'application/json')
        self.assertEqual(resp.header['Vary'], 'Cookie')
        self.assertEqual(resp.header['X-Frame-Options'], 'SAMEORIGIN')
        return

    def testUnavailableCRICHTTP(self):
        """
        Test HTTP exit code 503 - Service Unavailable for a CRIC response header
        when it also contains a HTTP string in the Set-Cookie header section
        """
        header = "HTTP/1.1 503 Service Unavailable\r\n" + self.HTTPheader

        resp = ResponseHeader(header)
        self.assertEqual(resp.status, 503)
        self.assertEqual(resp.reason, "Service Unavailable")
        self.assertFalse(resp.fromcache)
        self.assertIn("Content-Length", resp.header)
        self.assertIn("Date", resp.header)
        self.assertIn("Server", resp.header)
        self.assertIn("GRIDHTTP_PASSCODE", resp.header['Set-Cookie'])
        self.assertEqual(resp.header['Content-Type'], 'application/json')
        self.assertEqual(resp.header['Vary'], 'Cookie')
        self.assertEqual(resp.header['X-Frame-Options'], 'SAMEORIGIN')
        return

    def testHeadRequest(self):
        """
        Test a HEAD request.
        """
        params = {}
        headers = {}
        url = 'https://cmsweb-testbed.cern.ch/reqmgr2/data/info'
        res = self.mgr.getheader(url, params=params, headers=headers, ckey=self.ckey, cert=self.cert)
        self.assertEqual(res.getReason(), "OK")
        self.assertTrue(len(res.getHeader()) > 10)
        # Kubernetes cluster responds with a different Server header
        serverHeader = res.getHeaderKey("Server")
        self.assertTrue(serverHeader.startswith("nginx/") or serverHeader.startswith("CherryPy/") or serverHeader.startswith("openresty/"))

    def testHeadGzip(self):
        """
        Test a HEAD request incorrectly asking for gzip body
        """
        params = {}
        headers = {'Accept-Encoding': 'gzip'}
        url = 'https://cmsweb-testbed.cern.ch/reqmgr2/data/info'
        res = self.mgr.getdata(url, params=params, headers=headers, verb="HEAD",
                               ckey=self.ckey, cert=self.cert)
        self.assertEqual(res, "", "There is no body in HEAD requests")

        res = self.mgr.getheader(url, params=params, headers=headers, verb="HEAD",
                                 ckey=self.ckey, cert=self.cert)
        self.assertEqual(res.getReason(), "OK")
        self.assertEqual(res.getHeaderKey("Content-Encoding"), "gzip")
        # Kubernetes cluster responds with a different Server header
        serverHeader = res.getHeaderKey("Server")
        self.assertTrue(serverHeader.startswith("nginx/") or serverHeader.startswith("CherryPy/") or serverHeader.startswith("openresty/"))

    def testHeadUnsupportedAPI(self):
        """
        Test HEAD http request to an wrong endpoint (unsupported API),
        either with and without encoding compression (gzip)
        """
        params = {}
        url = 'https://cmsweb-testbed.cern.ch/reqmgr2/data/wrong_endpoint'
        for headers in [{}, {'Accept-Encoding': 'gzip'}]:
            try:
                res = self.mgr.getdata(url, params=params, headers=headers, verb="HEAD",
                                       ckey=self.ckey, cert=self.cert)
            except Exception as exc:
                self.assertTrue("404 Not Found" in str(exc))
            else:
                self.assertTrue("404 Not Found" in res)

            try:
                res = self.mgr.getheader(url, params=params, headers=headers, verb="HEAD",
                                         ckey=self.ckey, cert=self.cert)
            except Exception as exc:
                self.assertTrue("404 Not Found" in str(exc))
            else:
                self.assertEqual(res.getReason(), "Not Found")
                self.assertEqual(res.getHeaderKey("X-Error-Detail"), "API not supported")
                self.assertIsNone(res.getHeaderKey("Content-Encoding"))

    def testGetUnsupportedAPI(self):
        """
        Test GET http request to an wrong endpoint (unsupported API),
        either with and without encoding compression (gzip)
        """
        params = {}
        url = 'https://cmsweb-testbed.cern.ch/reqmgr2/data/wrong_endpoint'
        for headers in [{}, {'Accept-Encoding': 'gzip'}]:
            try:
                res = self.mgr.getdata(url, params=params, headers=headers, ckey=self.ckey, cert=self.cert)
            except Exception as exc:
                self.assertTrue("404 Not Found" in str(exc))
            else:
                self.assertTrue("404 Not Found" in res)

            try:
                res = self.mgr.getheader(url, params=params, headers=headers, ckey=self.ckey, cert=self.cert)
            except Exception as exc:
                self.assertTrue("404 Not Found" in str(exc))
            else:
                self.assertEqual(res.getReason(), "Not Found")
                self.assertEqual(res.getHeaderKey("X-Error-Detail"), "API not supported")
                self.assertIsNone(res.getHeaderKey("Content-Encoding"))

    def testToken(self):
        """
        Test setting up token header
        """
        iam_token = "eyJraWQiOiJyc2ExIiwiYWxnIjoiUlMyNTYifQ.eyJ3bGNnLnZlciI6IjEuMCIsInN1YiI6IjZjYTE3YzMyLTU0YWUtNDUzYy04YjQ1LTYyNTFkMjhlOTRhMSIsImF1ZCI6Imh0dHBzOlwvXC93bGNnLmNlcm4uY2hcL2p3dFwvdjFcL2FueSIsIm5iZiI6MTY0NTU1MDExMCwic2NvcGUiOiJhZGRyZXNzIHBob25lIG9wZW5pZCBvZmZsaW5lX2FjY2VzcyBwcm9maWxlIGVkdXBlcnNvbl9zY29wZWRfYWZmaWxpYXRpb24gZWR1cGVyc29uX2VudGl0bGVtZW50IGVtYWlsIHdsY2ciLCJpc3MiOiJodHRwczpcL1wvY21zLWF1dGgud2ViLmNlcm4uY2hcLyIsImV4cCI6MTY0NTU1MzcxMCwiaWF0IjoxNjQ1NTUwMTEwLCJqdGkiOiJkMWE3NzI0MC02Mjk5LTRiM2MtYmY5ZS0yOGNmYzgxY2ZjZDciLCJjbGllbnRfaWQiOiI0N2EwYjZkMC1mMzBlLTQ2OGItYmMwYy01MWVlNmE5Nzg2ODAifQ.Rpw6Mk_QzxtQVcbS-2OeVYGNMzyO08W540Sv3Nda2x3UJmPBRK4lnriRgSzMNTGP4y51rC5exbpf970MpJrjPaWqDhFWC--T6hxPAHhxHUxMTkXF5hUGXMLoNkCN5yR4eBJSAbgrgYJJwBcY5lMdCQ0BH5kQAUL6XRr3kvCrQQ0"
        os.environ['IAM_TOKEN'] = iam_token
        url = 'https://cmsweb-testbed.cern.ch'
        params = {}
        headers = {}
        try:
            # this call requires that our url (cmsweb-testbed) will
            # accept token based header, therefore we wrap it up in
            # try/except block to prevent from failing until
            # token based auth will be in place
            res = self.mgr.getheader(url, params=params, headers=headers, ckey=self.ckey, cert=self.cert)
            if isinstance(self.mgr.request_headers, dict):
                auth_header = self.mgr.request_headers.get('Authorization')
                self.assertTrue(auth_header != "", True)
            # the following test should be true only if we have valid token
            # TODO: so far I don't know how to test this since it requires
            # to obtain valid token
            # self.assertTrue(auth_header == 'Bearer {}'.format(iam_token), True)
        except Exception as exc:
            # do not use print since it cause E1601 ("print statement used"), see
            # https://github.com/PyCQA/pylint/issues/437
            traceback.print_exc()

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_SizeBased_

Suze based splitting algorithm that will produce a set of jobs for each file,
or a a set of files for each job.

"""




import logging

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.DataStructs.Fileset import Fileset
from WMCore.Services.UUIDLib import makeUUID

class SizeBased(JobFactory):
    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Implement size splitting algorithm.


        kwargs can take:
        size_per_job
        """
        sizePerJob = int(kwargs.get("size_per_job", 1000))
        locationDict = self.sortByLocation()

        for location in locationDict:
            self.newGroup()
            fileList     = locationDict[location]
            self.newJob(name = makeUUID())
            currentSize = 0

            for f in fileList:
                sizeOfFile = f['size']
                if sizeOfFile > sizePerJob:
                    if currentSize > 0:
                        self.newJob(name = makeUUID())
                    self.currentJob.addFile(f)
                    currentSize += sizeOfFile

                else:
                    if currentSize + sizeOfFile > sizePerJob:
                        #Create new jobs, because we are out of room
                        self.newJob(name = makeUUID())
                        currentSize = 0

                    if currentSize + sizeOfFile <= sizePerJob:

                        if not self.currentJob:
                            self.newJob(name = makeUUID())
                        #Add if it will be smaller
                        self.currentJob.addFile(f)
                        currentSize += sizeOfFile



"""
Created on Aug 13, 2014

@author: sryu
"""
from __future__ import (division, print_function)

from WMCore.ACDC.CouchService import CouchService
from WMCore.Lexicon import splitCouchServiceURL
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.Services.ReqMgrAux.ReqMgrAux import ReqMgrAux
from WMCore.Services.RequestDB.RequestDBReader import RequestDBReader


class CouchDBCleanup(CherryPyPeriodicTask):
    def __init__(self, rest, config):

        super(CouchDBCleanup, self).__init__(config)
        self.reqDB = RequestDBReader(config.reqmgrdb_url)
        # create another object to always retrieve up-to-date views (no stale)
        self.reqDBNoStale = RequestDBReader(config.reqmgrdb_url)
        self.reqDBNoStale._setNoStale()
        self.reqmgrAux = ReqMgrAux(config.reqmgr2_url, logger=self.logger)
        # statuses that we want to keep the transfer documents
        self.transferStatuses = ["assigned", "staging", "staged", "acquired",
                                 "failed", "running-open", "running-closed"]

        baseURL, acdcDB = splitCouchServiceURL(config.acdc_url)
        self.acdcService = CouchService(url=baseURL, database=acdcDB)

    def setConcurrentTasks(self, config):
        """
        sets the list of functions which
        """
        self.concurrentTasks = [{'func': self.acdcCleanup, 'duration': config.acdcCleanDuration},
                                {'func': self.auxCouchCleanup, 'duration': config.auxCleanDuration}]

    def auxCouchCleanup(self, config):
        """
        Cleanup TRANSFER documents from the reqmgr_auxiliary CouchDB.
        The list of status can be expanded in the future
        """
        self.logger.info("Fetching all the TRANSFER documents from CouchDB...")

        transferDocs = []
        for row in self.reqmgrAux.getTransferInfo("ALL_DOCS"):
            transferDocs.append(row['workflowName'])
        if not transferDocs:
            self.logger.info("  there are no transfer documents in the database.")
            return
        self.logger.info("%d transfer documents retrieved from the database.", len(transferDocs))

        self.logger.info("Fetching workflows from CouchDB for statuses: %s", self.transferStatuses)
        results = self.reqDBNoStale._getCouchView("bystatus", {}, self.transferStatuses)
        if not results["rows"]:
            self.logger.info("  there are no workflows matching those statuses.")
            return
        activeRequests = []
        for row in results["rows"]:
            activeRequests.append(row["id"])
        self.logger.info("%d workflows retrieved from the database.", len(activeRequests))

        # now find transfer documents (workflows) that are not active in the system
        docsToDelete = []
        for wflowName in transferDocs:
            if wflowName not in activeRequests:
                docsToDelete.append(wflowName)
        self.logger.info("Found %d transfer documents to delete", len(docsToDelete))

        for wflowName in docsToDelete:
            self.logger.info("Deleting transfer document: %s", wflowName)
            try:
                self.reqmgrAux.deleteConfigDoc("transferinfo", wflowName)
            except Exception as exc:
                self.logger.warning("Failed to delete transfer doc: %s. Error: %s", wflowName, str(exc))
        self.logger.info("Transfer documents cleanup completed.")

    def acdcCleanup(self, config):
        """
        gather active data statistics
        """
        self.logger.info("Fetching ACDC collection names...")
        originalRequests = self.acdcService.listCollectionNames()
        if not originalRequests:
            self.logger.info("  there are no collection documents to delete.")
            return

        # filter requests
        results = self.reqDB._getCouchView("byrequest", {}, originalRequests)
        # filter requests only in the following status
        deleteStates = ["announced", "rejected-archived", "aborted-archived", "normal-archived"]
        filteredRequests = []
        for row in results["rows"]:
            if row["value"][0] in deleteStates:
                filteredRequests.append(row["key"])

        total = 0
        for req in filteredRequests:
            try:
                self.logger.info("Removing ACDC collection for: %s", req)
                deleted = self.acdcService.removeFilesetsByCollectionName(req)
                if deleted is None:
                    self.logger.warning("  request '%s' already deleted", req)
                else:
                    total += len(deleted)
                    self.logger.info("ACDC collection %s deleted", req)
            except Exception as ex:
                self.logger.error("Failed to delete request: %s, will try again later. Error: %s", req, str(ex))
        self.logger.info("Total %s ACDC collections deleted", total)
        return



#!/usr/bin/env python
"""
_GetUnsubscribedBlocks_

Oracle implementation of GetUnsubscribedBlocks.
"""

from __future__ import division

from WMComponent.RucioInjector.Database.MySQL.GetUnsubscribedBlocks import GetUnsubscribedBlocks as MySQLBase


class GetUnsubscribedBlocks(MySQLBase):
    pass



"""
_Destroy_

Implementation of Destroy for Oracle

"""

from WMCore.Database.DBFormatter import DBFormatter

class Destroy(DBFormatter):

    def execute(self, subscription = None, conn = None, transaction = False):

        sql = """DECLARE
                 BEGIN

                   execute immediate 'purge recyclebin';

                   -- Tables
                   FOR o IN (SELECT table_name name FROM user_tables) LOOP
                     execute immediate 'drop table ' || o.name || ' cascade constraints';
                   END LOOP;

                   -- Sequences
                   FOR o IN (SELECT sequence_name name FROM user_sequences) LOOP
                     execute immediate 'drop sequence ' || o.name;
                   END LOOP;

                   -- Triggers
                   FOR o IN (SELECT trigger_name name FROM user_triggers) LOOP
                     execute immediate 'drop trigger ' || o.name;
                   END LOOP;

                   -- Synonyms
                   FOR o IN (SELECT synonym_name name FROM user_synonyms) LOOP
                     execute immediate 'drop synonym ' || o.name;
                   END LOOP;

                   -- Functions
                   FOR o IN (SELECT object_name name FROM user_objects WHERE object_type = 'FUNCTION') LOOP
                     execute immediate 'drop function ' || o.name;
                   END LOOP;

                   -- Procedures
                   FOR o IN (SELECT object_name name FROM user_objects WHERE object_type = 'PROCEDURE') LOOP
                     execute immediate 'drop procedure ' || o.name;
                   END LOOP;

                   execute immediate 'purge recyclebin';

                 END;"""

        self.dbi.processData(sql, {}, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_SetBlock_

MySQL implementation of DBSBufferFiles.SetBlock
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetBlock(DBFormatter):
    sql = """UPDATE dbsbuffer_file SET block_id =
               (SELECT id FROM dbsbuffer_block WHERE blockname = :block)
             WHERE dbsbuffer_file.lfn = :filelfn"""

    def execute(self, lfn, blockName, conn = None, transaction = None):
        """
        Insert block name into files

        """

        if isinstance(lfn, list):
            binds = []
            for entry in lfn:
                binds.append({'block': blockName, 'filelfn': entry})
        else:
            binds = {"block": blockName, "filelfn": lfn}


        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return



#!/usr/bin/env python
# encoding: utf-8
"""
TestInitCouchApp.py

Specialisation of TestInit for Tests that require a couch connection.

Own thing to avoid making entire Test stack depend on code from couchapp

Created by Dave Evans on 2010-08-19.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from builtins import object
from future import standard_library

standard_library.install_aliases()

import os
import urllib.parse

from couchapp.commands import push as couchapppush
from WMCore.Database.CMSCouch import CouchServer

from WMQuality.TestInit import TestInit


class CouchAppTestHarness(object):
    """
    Test Harness for installing a couch database instance with several couchapps
    in a unittest.setUp and wiping it out in a unittest.tearDown


    """

    def __init__(self, dbName, couchUrl=None):
        self.couchUrl = os.environ.get("COUCHURL", couchUrl)
        self.dbName = dbName
        if self.couchUrl == None:
            msg = "COUCHRURL env var not set..."
            raise RuntimeError(msg)
        if self.couchUrl.endswith('/'):
            raise RuntimeError("COUCHURL env var shouldn't end with /")
        self.couchServer = CouchServer(self.couchUrl)

    def create(self, dropExistingDb=True):
        """create couch db instance"""
        # import pdb
        # pdb.set_trace()
        if self.dbName in self.couchServer.listDatabases():
            if not dropExistingDb:
                return
            self.drop()

        self.couchServer.createDatabase(self.dbName)

    def drop(self):
        """blow away the couch db instance"""
        self.couchServer.deleteDatabase(self.dbName)

    def pushCouchapps(self, *couchappdirs):
        """
        push a list of couchapps to the database
        """
        for couchappdir in couchappdirs:
            couchapppush(couchappdir, "%s/%s" % (self.couchUrl, urllib.parse.quote_plus(self.dbName)))


class TestInitCouchApp(TestInit):
    """
    TestInit with baked in Couch goodness
    """

    def __init__(self, testClassName, dropExistingDb=True):
        TestInit.__init__(self, testClassName)
        self.databases = []
        self.couch = None
        # for experiments, after tests run, it's useful to have CouchDB
        # populated with the testing data - having tearDownCouch commented
        # out, this flag prevents from re-initializing the database
        self.dropExistingDb = dropExistingDb

    def couchAppRoot(self, couchapp):
        """Return parent path containing couchapp"""
        wmcoreroot = os.path.normpath(os.path.join(self.init.getWMBASE(), '..', '..', '..'))
        develPath = os.path.join(self.init.getWMBASE(), "src", "couchapps")
        if os.path.exists(os.path.join(develPath, couchapp)):
            return develPath
        elif os.path.exists(os.path.join(wmcoreroot, 'xdata', 'couchapps', couchapp)):
            return os.path.join(wmcoreroot, 'xdata', 'couchapps')
        elif os.path.exists(os.path.join(wmcoreroot, 'data', 'couchapps', couchapp)):
            return os.path.join(wmcoreroot, 'data', 'couchapps')
        raise OSError('Cannot find couchapp: %s' % couchapp)

    def setupCouch(self, dbName, *couchapps):
        """
        _setupCouch_

        Call in the setUp of your test to build a couch instance with the dbname provided
        and the required list of couchapps from WMCore/src/couchapps
        """
        self.databases.append(dbName)
        self.couch = CouchAppTestHarness(dbName)
        self.couch.create(dropExistingDb=self.dropExistingDb)
        # just create the db is couchapps are not specified
        if len(couchapps) > 0:
            self.couch.pushCouchapps(*[os.path.join(self.couchAppRoot(couchapp), couchapp) for couchapp in couchapps])

    couchUrl = property(lambda x: x.couch.couchUrl)
    couchDbName = property(lambda x: x.couch.dbName)

    def tearDownCouch(self):
        """
        _tearDownCouch_

        call this in tearDown to erase all evidence of your couch misdemeanours
        """
        for database in self.databases:
            couch = CouchAppTestHarness(database)
            couch.drop()

        self.couch = None
        return



#!/usr/bin/env python

"""
Oracle implementation of GetChecksum
"""





from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetChecksum import GetChecksum as MySQLGetChecksum

class GetChecksum(MySQLGetChecksum):
    """
    Identical to MySQL Version

    """



#!/usr/bin/env python

import random

from WMCore.FwkJobReport import Report
from WMCore.DataStructs.Run import Run
from WMCore.Services.UUIDLib import makeUUID

outputModules = ["outputModule1", "outputModule2", "outputModule3",
                 "outputModule4", "outputModule5", "outputModule6",
                 "outputModule7", "outputModule8", "outputModule9",
                 "outputModule10"]

runInfo = Run(1)
runInfo.extendLumis([11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
                     25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38,
                     39, 40])

totalReports = 25
inputFilesPerReport = 50

inputFileCounter = 0
for i in range(totalReports):
    loadTestReport = Report.Report("cmsRun1")
    loadTestReport.addInputSource("PoolSource")

    for j in range(inputFilesPerReport):
        inputFile = loadTestReport.addInputFile("PoolSource", lfn = "input%i" % inputFileCounter,
                                                events = 600000, size = 600000)
        inputFileCounter += 1

    Report.addRunInfoToFile(inputFile, runInfo)

    for outputModule in outputModules:
        loadTestReport.addOutputModule(outputModule)
        datasetInfo = {"applicationName": "cmsRun", "applicationVersion": "CMSSW_3_3_5_patch3",
                       "primaryDataset": outputModule, "dataTier": "RAW",
                       "processedDataset": "LoadTest10"}
        fileAttrs = {"lfn": makeUUID(), "location": "cmssrm.fnal.gov",
                     "checksums": {"adler32": "ff810ec3", "cksum": "2212831827"},
                     "events": random.randrange(500, 5000, 50),
                     "merged": True,
                     "size": random.randrange(1000, 2000, 100000000),
                     "module_label": outputModule, "dataset": datasetInfo}

        outputFile = loadTestReport.addOutputFile(outputModule, fileAttrs)
        Report.addRunInfoToFile(outputFile, runInfo)

    loadTestReport.persist("HeritageTest%02d.pkl" % i)



"""
File       : MSOtput.py

Description: MSOutput.py class provides the whole logic behind
the Output data placement in WMCore MicroServices.
"""

# futures
from __future__ import division, print_function
from future.utils import viewitems

# system modules
import time
from pymongo import IndexModel, ReturnDocument, errors
from pprint import pformat
from threading import current_thread
from retry import retry

# WMCore modules
from WMCore.MicroService.DataStructs.DefaultStructs import OUTPUT_REPORT
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.MicroService.Tools.Common import gigaBytes
from WMCore.Services.CRIC.CRIC import CRIC
from WMCore.Services.DBS.DBS3Reader import getDataTiers
from Utils.Pipeline import Pipeline, Functor
from WMCore.Database.MongoDB import MongoDB
from WMCore.MicroService.MSOutput.MSOutputTemplate import MSOutputTemplate
from WMCore.MicroService.MSOutput.RelValPolicy import RelValPolicy
from WMCore.WMException import WMException


class MSOutputException(WMException):
    """
    General Exception Class for MSOutput Module in WMCore MicroServices
    """
    def __init__(self, message):
        self.myMessage = "MSOtputException: %s" % message
        super(MSOutputException, self).__init__(self.myMessage)


class EmptyResultError(MSOutputException):
    """
    A MSOutputException signalling an empty result from database query.
    """
    def __init__(self, message=None):
        if message:
            self.myMessage = "EmptyResultError: %s"
        else:
            self.myMessage = "EmptyResultError."
        super(EmptyResultError, self).__init__(self.myMessage)


class UnsupportedError(MSOutputException):
    """
    A MSOutputException signalling an unsupported mode for a function or method.
    """
    def __init__(self, message=None):
        if message:
            self.myMessage = "UnsupportedError: %s"
        else:
            self.myMessage = "UnsupportedError."
        super(UnsupportedError, self).__init__(self.myMessage)


class MSOutput(MSCore):
    """
    MSOutput.py class provides the whole logic behind the Output data placement
    in MicroServices.
    """

    def __init__(self, msConfig, mode, logger=None):
        """
        Runs the basic setup and initialization for the MSOutput module
        :microConfig: microservice configuration
        :mode: MSOutput Run mode:
            - MSOutputConsumer:
                Reads The workflow and transfer subscriptions from MongoDB and
                makes transfer subscriptions.
            - MSOutputProducer:
                Fetches Workflows in a given status from Reqmgr2 then creates
                and uploads the documents to MongoDB.
        """
        super(MSOutput, self).__init__(msConfig, logger=logger)

        self.mode = mode
        self.msConfig.setdefault("limitRequestsPerCycle", 500)
        self.msConfig.setdefault("enableDataPlacement", False)
        self.msConfig.setdefault("enableRelValCustodial", False)
        self.msConfig.setdefault("excludeDataTier", [])
        self.msConfig.setdefault("rucioAccount", 'wmcore_transferor')
        self.msConfig.setdefault("rucioRSEAttribute", 'ddm_quota')
        self.msConfig.setdefault("rucioDiskRuleWeight", 'ddm_quota')
        self.msConfig.setdefault("rucioTapeExpression", 'rse_type=TAPE\cms_type=test')
        # This Disk expression wil target all real DISK T1 and T2 RSEs
        self.msConfig.setdefault("rucioDiskExpression", '(tier=2|tier=1)&cms_type=real&rse_type=DISK')
        # fetch documents created in the last 6 months (default value)
        self.msConfig.setdefault("mongoDocsCreatedSecs", 6 * 30 * 24 * 60 * 60)
        self.msConfig.setdefault("sendNotification", False)
        self.msConfig.setdefault("relvalPolicy", [])
        self.msConfig.setdefault("ruleLifetimeRelVal", [])

        self.uConfig = {}
        # service name used to route alerts via AlertManager
        self.alertServiceName = "ms-output"

        # RelVal output data placement policy from the service configuration
        self.msConfig.setdefault("dbsUrl", "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader")
        allDBSDatatiers = getDataTiers(self.msConfig['dbsUrl'])
        allDiskRSEs = self.rucio.evaluateRSEExpression("*", returnTape=False)
        self.relvalPolicy = RelValPolicy(self.msConfig['relvalPolicy'],
                                         self.msConfig['ruleLifetimeRelVal'],
                                         allDBSDatatiers, allDiskRSEs, logger=logger)

        self.cric = CRIC(logger=self.logger)
        self.uConfig = {}
        self.campaigns = {}
        self.psn2pnnMap = {}

        self.msConfig.setdefault("mongoDBRetryCount", 3)
        self.msConfig.setdefault("mongoDBReplicaset", None)
        self.msConfig.setdefault("mockMongoDB", False)

        msOutIndex = IndexModel('RequestName', unique=True)
        msOutDBConfig = {
            'database': self.msConfig['mongoDB'],
            'server': self.msConfig['mongoDBUrl'],
            'port': self.msConfig['mongoDBPort'],
            'logger': self.logger,
            'create': True,
            'replicaset': self.msConfig['mongoDBReplicaset'],
            'collections': [
                ('msOutRelValColl', msOutIndex),
                ('msOutNonRelValColl', msOutIndex)]}

        mongoClt = MongoDB(**msOutDBConfig)
        self.msOutDB = getattr(mongoClt, self.msConfig['mongoDB'])
        self.msOutRelValColl = self.msOutDB['msOutRelValColl']
        self.msOutNonRelValColl = self.msOutDB['msOutNonRelValColl']
        self.currThread = None
        self.currThreadIdent = None


    @retry(tries=3, delay=2, jitter=2)
    def updateCaches(self):
        """
        Fetch some data required for the output logic, e.g.:
        * unified configuration
        """
        self.logger.info("%s: Updating local cache information.", self.currThreadIdent)

        self.uConfig = self.unifiedConfig()
        campaigns = self.reqmgrAux.getCampaignConfig("ALL_DOCS")
        self.psn2pnnMap = self.cric.PSNtoPNNMap()
        if not self.uConfig:
            raise RuntimeWarning("Failed to fetch the unified configuration")
        elif not campaigns:
            raise RuntimeWarning("Failed to fetch the campaign configurations")
        elif not self.psn2pnnMap:
            raise RuntimeWarning("Failed to fetch PSN x PNN map from CRIC")
        else:
            # let's make campaign look-up easier and more efficient
            self.campaigns = {}
            for camp in campaigns:
                self.campaigns[camp['CampaignName']] = camp

    def execute(self, reqStatus):
        """
        Executes the whole output data placement logic. However, updating the
        local caches is a requirement to proceed with the rest of the execution.
        :return: summary report for an execution cycle
        """
        summary = dict(OUTPUT_REPORT)

        self.currThread = current_thread()
        self.currThreadIdent = self.currThread.name
        self.updateReportDict(summary, "thread_id", self.currThreadIdent)

        try:
            self.updateCaches()
        except RuntimeWarning as ex:
            msg = "{}: All retries exhausted! Last error was: '{}'".format(self.currThreadIdent,
                                                                           str(ex))
            msg += "\nRetrying to update caches again in the next cycle."
            self.logger.error(msg)
            self.updateReportDict(summary, "error", msg)
            return summary
        except Exception as ex:
            msg = "{}: Unknown exception updating caches. ".format(self.currThreadIdent)
            msg += "Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)
            return summary

        if self.mode == 'MSOutputProducer':
            self._executeProducer(reqStatus, summary)

        elif self.mode == 'MSOutputConsumer':
            self._executeConsumer(summary)

        else:
            msg = "MSOutput is running in unsupported mode: %s\n" % self.mode
            msg += "Skipping the current run!"
            self.logger.warning(msg)
            self.updateReportDict(summary, "error", msg)

        return summary

    def _executeProducer(self, reqStatus, summary):
        """
        The function to update caches and to execute the Producer function itself
        :param summary: dictionary with some high level summary for this cycle execution
        """
        msg = "{}: MSOutput is running in mode: {}".format(self.currThreadIdent, self.mode)
        self.logger.info(msg)

        try:
            mongoDocNames = self.getRecentDocNamesFromMongo()
        except Exception as err:  # general error
            mongoDocNames = []
            msg = "{}: Unknown exception while fetching documents from MongoDB. ".format(self.currThreadIdent)
            msg += "Error: {}".format(str(err))
            self.logger.exception(msg)

        try:
            requestRecords = {}
            for status in reqStatus:
                requestRecords.update(self.getRequestRecords(status))
        except Exception as err:  # general error
            msg = "{}: Unknown exception while fetching requests from ReqMgr2. ".format(self.currThreadIdent)
            msg += "Error: {}".format(str(err))
            self.logger.exception(msg)

        # filter out documents already produced
        finalRequests = []
        for reqName, reqData in viewitems(requestRecords):
            if reqName in mongoDocNames:
                self.logger.info("Mongo document already created for %s, skipping it.", reqName)
            else:
                finalRequests.append(reqData)
        msg = "Retrieved {} recent docs from MongoDB, ".format(len(mongoDocNames))
        msg += "{} requests from ReqMgr2, and {} are new requests to be processed.".format(len(requestRecords),
                                                                                           len(finalRequests))
        self.logger.info(msg)

        try:
            total_num_requests = self.msOutputProducer(finalRequests)
            msg = "{}: Total {} requests processed from the streamer. ".format(self.currThreadIdent,
                                                                               total_num_requests)
            self.logger.info(msg)
            self.updateReportDict(summary, "total_num_requests", total_num_requests)
        except Exception as ex:
            msg = "{}: Unknown exception while running the Producer thread. ".format(self.currThreadIdent)
            msg += "Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

    def _executeConsumer(self, summary):
        """
        The function to execute the Consumer function itself
        :param summary: dictionary with some high level summary for this cycle execution
        """
        msg = "{}: MSOutput is running in mode: {} ".format(self.currThreadIdent, self.mode)
        self.logger.info(msg)
        msg = "{}: Service set to process up to {} requests ".format(self.currThreadIdent,
                                                                     self.msConfig["limitRequestsPerCycle"])
        msg += "per cycle per each type 'RelVal' and 'NonRelval' workflows."
        self.logger.info(msg)

        if not self.msConfig['enableDataPlacement']:
            msg = "{} enableDataPlacement = False. ".format(self.currThreadIdent)
            msg += "Running the MSOutput service in dry run mode"
            self.logger.warning(msg)

        try:
            total_num_requests = self.msOutputConsumer()
            msg = "{}: Total {} requests processed. ".format(self.currThreadIdent,
                                                             total_num_requests)
            self.logger.info(msg)
            self.updateReportDict(summary, "total_num_requests", total_num_requests)
        except Exception as ex:
            msg = "{}: Unknown exception while running Consumer thread. ".format(self.currThreadIdent)
            msg += "Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

    def makeSubscriptions(self, workflow):
        """
        The common function to make the final subscriptions
        :param workflow: a MSOutputTemplate object workflow
        :return: the MSOutputTemplate object itself (with the necessary updates in place)
        """
        if not isinstance(workflow, MSOutputTemplate):
            msg = "Unsupported type object '{}' for workflows! ".format(type(workflow))
            msg += "It needs to be of type: MSOutputTemplate"
            raise UnsupportedError(msg)

        # NOTE:
        #    Here is just an example construction of the function. None of the
        #    data structures used to visualise it is correct. To Be Updated

        ruleAttrs = {'activity': 'Production Output',
                     'lifetime': self.msConfig['rulesLifetime'],
                     'account': self.msConfig['rucioAccount'],
                     'grouping': "ALL",
                     'comment': 'WMCore MSOutput output data placement'}
        # add a configurable weight value
        ruleAttrs["weight"] = self.msConfig['rucioDiskRuleWeight']

        # if anything fail along the way, set it back to "pending"
        transferStatus = "done"
        for dMap in workflow['OutputMap']:
            if dMap['Copies'] == 0:
                msg = "Output dataset configured to 0 copies, so skipping it. Details:"
                msg += "\n\tWorkflow name: {}".format(workflow['RequestName'])
                msg += "\n\tDataset name: {}".format(dMap['Dataset'])
                msg += "\n\tCampaign name: {}".format(dMap['Campaign'])
                self.logger.warning(msg)
                continue
            if dMap['DiskRuleID']:
                msg = "Output dataset: {} from workflow: {} ".format(dMap['Dataset'], workflow['RequestName'])
                msg += " has been already locked by rule id: {}".format(dMap['DiskRuleID'])
                self.logger.info(msg)
                continue

            ruleAttrs.update({'copies': dMap['Copies']})
            # RelVals have a different lifetime setting and it depends on the sample type
            if workflow['IsRelVal']:
                ruleLifeT = self.relvalPolicy.getLifetimeByDataset(dMap['Dataset'])
                ruleAttrs["lifetime"] = ruleLifeT

            msg = "Performing rucio rule creation for workflow: {}, ".format(workflow['RequestName'])
            msg += "with the following information: {}".format(dMap)
            self.logger.info(msg)

            if self.msConfig['enableDataPlacement']:
                resp = self.rucio.createReplicationRule(dMap['Dataset'], dMap['DiskDestination'], **ruleAttrs)
                if not resp:
                    # then the call failed
                    transferStatus = "pending"
                elif len(resp) == 1:
                    dMap['DiskRuleID'] = resp[0]
                elif len(resp) > 1:
                    msg = "Rule creation returned multiple rule IDs and it needs to be investigated!!! "
                    msg += "For DID: {}, rseExpr: {} and rucio account: {}".format(dMap['Dataset'],
                                                                                   dMap['DiskDestination'],
                                                                                   ruleAttrs['account'])
                    self.logger.critical(msg)
                    return workflow
            else:
                msg = "DRY-RUN RUCIO: skipping rule creation for DID: {}, ".format(dMap['Dataset'])
                msg += "rseExpr: {} and standard parameters: {}".format(dMap['DiskDestination'], ruleAttrs)
                self.logger.info(msg)

        # Finally, update the MSOutput template document with either partial or
        # complete transfer ids
        self.docKeyUpdate(workflow, OutputMap=workflow['OutputMap'])
        workflow.updateTime()
        if transferStatus == "done":
            self.logger.info("All the disk requests succeeded for: %s. Marking it as 'done'",
                             workflow['RequestName'])
            self.docKeyUpdate(workflow, TransferStatus='done')
        else:
            self.logger.info("Disk requests partially successful for: %s. Keeping it 'pending'",
                             workflow['RequestName'])

        # NOTE:
        #    if we are about to implement this through a pipeline we MUST not
        #    return the result here but the WHOLE document with updated fields
        #    for the transfer as it will be passed to the next function in
        #    the pipeline and uploaded to MongoDB
        return workflow

    def makeTapeSubscriptions(self, workflow):
        """
        Makes the output data placement to the Tape endpoints. It works either with
        PhEDEx or with Rucio, configurable. It also relies on the Unified configuration
        to decide whether a given datatier can go to tape, and where it can be auto-approved.
        :param workflow: a MSOutputTemplate object representing a workflow
        :return: the MSOutputTemplate object itself (with the necessary updates in place)
        """
        # if anything fails along the way, set it back to "pending"
        transferStatus = "done"

        # this RSE name will be used for all output datasets to be subscribed
        # within this workflow
        dataBytesForTape = self._getDataVolumeForTape(workflow)
        tapeRSE, requiresApproval = self._getTapeDestination(dataBytesForTape)
        self.logger.info("Workflow: %s, total output size: %s GB, against RSE: %s",
                         workflow['RequestName'], gigaBytes(dataBytesForTape), tapeRSE)
        for dMap in workflow['OutputMap']:
            if not self.canDatasetGoToTape(dMap, workflow):
                continue

            # this RSE name will be used for all output datasets to be subscribed
            # within this workflow
            dMap['TapeDestination'] = tapeRSE
            ruleAttrs = {'activity': 'Production Output',
                         'account': self.msConfig['rucioAccount'],
                         'copies': 1,
                         'grouping': "ALL",
                         'ask_approval': requiresApproval,
                         'comment': 'WMCore MSOutput output data placement'}
            msg = "Creating Rucio TAPE rule for container: {} and RSE: {}".format(dMap['Dataset'],
                                                                                  dMap['TapeDestination'])
            self.logger.info(msg)

            if self.msConfig['enableDataPlacement']:
                resp = self.rucio.createReplicationRule(dMap['Dataset'], dMap['TapeDestination'], **ruleAttrs)
                if not resp:
                    # then the call failed
                    transferStatus = "pending"
                elif len(resp) == 1:
                    dMap['TapeRuleID'] = resp[0]
                elif len(resp) > 1:
                    msg = "Tape rule creation returned multiple rule IDs and it needs to be investigated!!! "
                    msg += "For DID: {}, rseExpr: {} and rucio account: {}".format(dMap['Dataset'],
                                                                                   dMap['TapeDestination'],
                                                                                   ruleAttrs['account'])
                    self.logger.critical(msg)
                    return workflow
            else:
                msg = "DRY-RUN RUCIO: skipping tape rule creation for DID: {}, ".format(dMap['Dataset'])
                msg += "rseExpr: {} and standard parameters: {}".format(dMap['TapeDestination'], ruleAttrs)
                self.logger.info(msg)

        # Finally, update the MSOutput template document with either partial or
        # complete transfer ids
        self.docKeyUpdate(workflow, OutputMap=workflow['OutputMap'])
        workflow.updateTime()
        # NOTE: updating the TransferStatus at this stage is a bit trickier, we
        # cannot bypass bad disk data placements!
        if transferStatus == "done" and workflow['TransferStatus'] == "done":
            self.logger.info("All the tape requests succeeded for: %s. Marking it as 'done'",
                             workflow['RequestName'])
        elif transferStatus == "done" and workflow['TransferStatus'] == "pending":
            self.logger.info("All the tape requests succeeded for: %s, but disk ones are still pending",
                             workflow['RequestName'])
        elif transferStatus == "pending" and workflow['TransferStatus'] == "done":
            self.logger.info("Tape requests partially successful for: %s. Marking it as 'pending'",
                             workflow['RequestName'])
            self.docKeyUpdate(workflow, TransferStatus='pending')
        else:
            self.logger.info("Tape requests partially successful for: %s. Keeping it as 'pending'",
                             workflow['RequestName'])

        return workflow

    def _getTapeDestination(self, dataSize):
        """
        Depending on which Data Management system this service is configured
        to use. Run a different procedure to find out which tape endpoint to
        select as a destination for all the output datasets in a given workflow
        :param dataSize: integer with the total amount of data to be transferred, in bytes
        :return: a string with the RSE name
        """
        # This API returns a tuple with the RSE name and whether it requires approval
        return self.rucio.pickRSE(rseExpression=self.msConfig["rucioTapeExpression"],
                                  rseAttribute=self.msConfig["rucioRSEAttribute"],
                                  minNeeded=dataSize)

    def getRequestRecords(self, reqStatus):
        """
        Queries ReqMgr2 for requests in a given status.
        NOTE: to be taken from MSTransferor with minor changes
        """
        self.logger.info("Fetching requests in status: %s", reqStatus)
        result = self.reqmgr2.getRequestByStatus([reqStatus], detail=True)
        if not result:
            requests = {}
        else:
            requests = result[0]
        self.logger.info('  retrieved %s requests in status: %s', len(requests), reqStatus)

        return requests

    def msOutputConsumer(self):
        """
        A top level function to drive the creation and book keeping of all the
        subscriptions to the Data Management System
        """
        # DONE:
        #    Done: To check if the 'enableDataPlacement' flag is really taken into account
        #    Done: To make this for both relvals and non relvals
        #    Done: To return the result
        #    Done: To make report document
        #    Done: To build it through a pipe
        #    Done: To write back the updated document to MonogoDB
        msPipelineRelVal = Pipeline(name="MSOutputConsumer PipelineRelVal",
                                    funcLine=[Functor(self.makeSubscriptions),
                                              Functor(self.makeTapeSubscriptions),
                                              Functor(self.docUploader,
                                                      update=True,
                                                      keys=['LastUpdate',
                                                            'TransferStatus',
                                                            'OutputMap']),
                                              Functor(self.docDump, pipeLine='PipelineRelVal'),
                                              Functor(self.docCleaner)])
        msPipelineNonRelVal = Pipeline(name="MSOutputConsumer PipelineNonRelVal",
                                       funcLine=[Functor(self.makeSubscriptions),
                                                 Functor(self.makeTapeSubscriptions),
                                                 Functor(self.docUploader,
                                                         update=True,
                                                         keys=['LastUpdate',
                                                               'TransferStatus',
                                                               'OutputMap']),
                                                 Functor(self.docDump, pipeLine='PipelineNonRelVal'),
                                                 Functor(self.docCleaner)])

        wfCounterTotal = 0
        mQueryDict = {'TransferStatus': 'pending'}
        pipeCollections = [(msPipelineRelVal, self.msOutRelValColl),
                           (msPipelineNonRelVal, self.msOutNonRelValColl)]
        for pipeColl in pipeCollections:
            wfCounters = 0
            pipeLine = pipeColl[0]
            dbColl = pipeColl[1]
            pipeLineName = pipeLine.getPipelineName()
            for docOut in self.getDocsFromMongo(mQueryDict, dbColl, self.msConfig['limitRequestsPerCycle']):
                # FIXME:
                #    To redefine those exceptions as MSoutputExceptions and
                #    start using those here so we do not mix with general errors
                try:
                    pipeLine.run(docOut)
                except (KeyError, TypeError) as ex:
                    msg = "%s Possibly malformed record in MongoDB. Err: %s. " % (pipeLineName, str(ex))
                    msg += "Continue to the next document."
                    self.logger.exception(msg)
                    continue
                except EmptyResultError:
                    msg = "%s All relevant records in MongoDB exhausted. " % pipeLineName
                    msg += "We are done for the current cycle."
                    self.logger.info(msg)
                    break
                except Exception as ex:
                    msg = "%s General error from pipeline. Err: %s. " % (pipeLineName, str(ex))
                    msg += "Will retry again in the next cycle."
                    self.logger.exception(msg)
                    break
                wfCounters += 1
            self.logger.info("Processed %d workflows from pipeline: %s", wfCounters, pipeLineName)
            wfCounterTotal += wfCounters

        return wfCounterTotal

    def msOutputProducer(self, requestRecords):
        """
        A top level function to fetch requests from ReqMgr2, and produce the correspondent
        records for MSOutput in MongoDB.
        :param requestRecords: list of request dictionaries retrieved from ReqMgr2

        It's implemented as a pipeline, performing the following sequential actions:
           1) document transformer - creates a MSOutputTemplate object from the request dict
           2) document info updater - parses the MSOutputTemplate object and updates the
              necessary data structure mapping output/locations/campaign/etc
           3) document uploader - inserts the MSOutputTemplate object into the correct
              MongoDB collection (ReVal is separated from standard workflows)
           4) document cleaner - releases memory reference to the MSOutputTemplate object
        """
        # DONE:
        #    to set a destructive function at the end of the pipeline
        # NOTE:
        #    To discuss the collection names
        # NOTE:
        #    Here we should never use docUploader with `update=True`, because
        #    this will erase the latest state of already existing and fully or
        #    partially processed documents by the Consumer pipeline
        self.logger.info("Running the msOutputProducer ...")
        msPipeline = Pipeline(name="MSOutputProducer Pipeline",
                              funcLine=[Functor(self.docTransformer),
                                        Functor(self.docInfoUpdate),
                                        Functor(self.docUploader),
                                        Functor(self.docCleaner)])
        # TODO:
        #    To generate the object from within the Function scope see above.
        counter = 0
        for request in requestRecords:
            counter += 1
            pipeLineName = msPipeline.getPipelineName()
            try:
                msPipeline.run(request)
            except (KeyError, TypeError) as ex:
                msg = "%s Possibly broken read from ReqMgr2 API or other. Err: %s." % (pipeLineName, str(ex))
                msg += " Continue to the next document."
                self.logger.exception(msg)
                continue
            except Exception as ex:
                msg = "%s General Error from pipeline. Err: %s. " % (pipeLineName, str(ex))
                msg += "Giving up Now."
                self.logger.exception(str(ex))
                break
        return counter

    def docTransformer(self, doc):
        """
        A function used to transform a request record from reqmgr2 to a document
        suitable for uploading to Mongodb
        """
        # Solution 1: Destructive function - to force clear of the the externally
        #             referenced object and to return a new one (current solution)
        #             NOTE: Leaves an empty dictionary behind (the clear method just
        #                   clears all the keys of the dict, but does not delete it)
        # Solution 2: To work in place (will keep the dynamic structure of the passed dict)
        # Solution 3: To have 2 object buffers for the two diff types outside the function
        try:
            msOutDoc = MSOutputTemplate(doc)
        except Exception as ex:
            msg = "ERR: Unable to create MSOutputTemplate for document: \n%s\n" % pformat(doc)
            msg += "ERR: %s" % str(ex)
            self.logger.exception(msg)
            raise ex
        return msOutDoc

    def docDump(self, msOutDoc, pipeLine=None):
        """
        Prints document contents
        """
        msg = "{}: {}: Processed 'msOutDoc' with '_id': {}.".format(self.currThreadIdent,
                                                                    pipeLine,
                                                                    msOutDoc['_id'])
        self.logger.info(msg)
        self.logger.debug(pformat(msOutDoc))
        return msOutDoc

    def docKeyUpdate(self, msOutDoc, **kwargs):
        """
        A function used to update one or few particular fields in a document
        :**kwargs: The keys/value pairs to be updated (will be tested against MSOutputTemplate)
        """
        for key, value in viewitems(kwargs):
            try:
                msOutDoc.setKey(key, value)
                msOutDoc.updateTime()
            except Exception as ex:
                msg = "Cannot update key {} for doc: {}\n".format(key, msOutDoc['_id'])
                msg += "Error: {}".format(str(ex))
                self.logger.warning(msg)
        return msOutDoc

    def docInfoUpdate(self, msOutDoc):
        """
        Parses the request parameters (a mongoDB record, not yet persisted) and finds
        out what are the disk destinations and how many copies of each dataset need to
        be made.
        :param msOutDoc: a MSOutput template object
        :return: nothing, the MSOutput template record is update in memory.
        """
        self.logger.info("Producing MongoDB record for workflow: %s", msOutDoc["RequestName"])
        updatedOutputMap = []
        notFoundDIDs = []
        for dataItem in msOutDoc['OutputMap']:
            if msOutDoc['RequestType'] == "Resubmission":
                # make sure not to subscribe the same datasets multiple times, even
                # worse, to different locations! Initial workflow will take care of everything!
                dataItem['Copies'] = 0
                updatedOutputMap.append(dataItem)
                continue
            # Fetch the dataset size, even if it does not go to Disk (it might go to Tape)
            try:
                bytesSize = self._getDatasetSize(dataItem['Dataset'])
            except KeyError:
                # then this container is unknown to Rucio, bypass and make an alert
                # Error is already reported in the Rucio module, do not spam here!
                dataItem['DatasetSize'] = 0
                dataItem['Copies'] = 0
                updatedOutputMap.append(dataItem)
                notFoundDIDs.append(dataItem['Dataset'])
                continue

            dataItem['DatasetSize'] = bytesSize

            if not self.canDatasetGoToDisk(dataItem, msOutDoc['IsRelVal']):
                # nope, this dataset cannot proceed to Disk!!
                dataItem['Copies'] = 0
                updatedOutputMap.append(dataItem)
                continue

            try:
                dataItem['Copies'] = self.campaigns[dataItem['Campaign']]["MaxCopies"]
            except KeyError:
                # it can happen for RelVals, but canDatasetGoToDisk method above
                # will already take the necessary action for non existent campaign
                dataItem['Copies'] = 1

            if msOutDoc['IsRelVal']:
                destination = self.relvalPolicy.getDestinationByDataset(dataItem['Dataset'])
                if destination:
                    # ensure each RelVal destination gets a copy of the data
                    dataItem['Copies'] = len(destination)
                    dataItem['DiskDestination'] = '|'.join(destination)
                else:
                    self.logger.warning("RelVal dataset: %s without any destination", dataItem['Dataset'])
                    dataItem['Copies'] = 0
                    updatedOutputMap.append(dataItem)
                    continue
            else:
                dataItem['DiskDestination'] = self.msConfig["rucioDiskExpression"]
            updatedOutputMap.append(dataItem)

        # if there were containers not found in Rucio, create an email alert
        if notFoundDIDs:
            # log and send alert via AlertManager API
            self.alertDIDNotFound(msOutDoc["RequestName"], notFoundDIDs)

        try:
            msOutDoc.updateDoc({"OutputMap": updatedOutputMap}, throw=True)
        except Exception as ex:
            msg = "%s: Could not update the additional information for "
            msg += "'msOutDoc' with '_id': %s \n"
            msg += "Error: %s"
            self.logger.exception(msg, self.currThreadIdent, msOutDoc['_id'], str(ex))
        return msOutDoc

    def _getDatasetSize(self, datasetName):
        """
        Retrieve the dataset size from the correct DM system
        This size is needed for the tape data placement
        :param datasetName: string with the dataset name
        :return: an integer with the total dataset size, in bytes
        """
        didInfo = self.rucio.getDID(datasetName)
        # let the exception be raised if we failed to calculate the dataset size
        return didInfo["bytes"]

    def canDatasetGoToDisk(self, dataItem, isRelVal=False):
        """
        This function evaluates whether a dataset can be passed to the
        Data Management system, considering the following configurations:
          1) list of blacklisted tiers in the MicroService configuration
          2) list of white listed tiers bypassing the Unified configuration
          3) list of black and white listed tiers in the Unified config
        :param dataItem: dictionary information for this dataset, from MongoDB record
        :param isRelVal: boolean flag identifying if dataset belongs to a RelVal request
        :return: True if the dataset is allowed to pass, False otherwise
        """
        # Bypass every configuration for RelVals, keep everything on disk
        if isRelVal:
            return True
        dataTier = dataItem['Dataset'].split('/')[-1]
        if dataTier in self.msConfig['excludeDataTier']:
            self.logger.warning("Skipping dataset: %s because it's excluded in the MS configuration",
                                dataItem['Dataset'])
            return False

        try:
            if dataTier in self.campaigns[dataItem['Campaign']]["TiersToDM"]:
                return True
        except KeyError:
            if isRelVal:
                msg = "Campaign not found for RelVal dataset: {} ".format(dataItem['Dataset'])
                msg += "under campaign: {}. Letting it pass though...".format(dataItem['Campaign'])
                self.logger.warning(msg)
                return True
            # log and send alert via AlertManager API
            self.alertCampaignNotFound(dataItem['Campaign'], dataItem['Dataset'])
            raise

        if dataTier in self.uConfig['tiers_to_DDM']['value']:
            return True
        elif dataTier in self.uConfig['tiers_no_DDM']['value']:
            return False
        else:
            # log and send alert via AlertManager API
            self.alertDatatierNotFound(dataTier, dataItem['Dataset'], isRelVal)
            return True

    def _getDataVolumeForTape(self, workflow):
        """
        This function does a similar logic as `canDatasetGoToTape` and
        calculates the total size of all the output datasets that need
        to be pinned on tape
        :param workflow: MSOutputTemplate object retrieved from MongoDB
        :return: integer with the total size in bytes
        """
        totalSize = 0
        for dataItem in workflow['OutputMap']:
            if workflow['IsRelVal'] and not self.msConfig['enableRelValCustodial']:
                return False
            if dataItem['TapeRuleID']:
                continue
            dataTier = dataItem['Dataset'].split('/')[-1]
            if dataTier in self.msConfig['excludeDataTier']:
                continue
            elif dataTier in self.uConfig['tiers_with_no_custodial']['value']:
                continue

            # otherwise, we are about to transfer it
            totalSize += dataItem['DatasetSize']

        return totalSize

    def canDatasetGoToTape(self, dataItem, workflow):
        """
        This function evaluates whether a container can be passed to the
        Data Management system, considering the following configurations:
          1) list of tiers banned in the MicroService configuration
          2) list of tiers allowed to bypass the Unified configuration
          3) list of allowed and banned tiers in the Unified config
        :param dataItem: output map dictionary present in the MongoDB record
        :param workflow: MSOutputTemplate object retrieved from MongoDB
        :return: True if the dataset is allowed to pass, False otherwise

        NOTE: changes here should usually be applied to `_getDataVolumeForTape` too
        """
        # NOTE: Unified has a `tape_size_limit` parameter, to prevent automatic
        # tape subscription for too large samples. We are not going to implement
        # it - for the moment - at least.
        if workflow['IsRelVal'] and not self.msConfig['enableRelValCustodial']:
            return False
        if workflow['RequestType'] == "Resubmission":
            # their parent/original workflow will take care of all the data placement
            return False
        if int(dataItem['DatasetSize']) == 0:
            # container hasn't been produced in the workflow
            return False

        if dataItem['TapeRuleID']:
            msg = "Output dataset: {} from workflow: {} ".format(dataItem['Dataset'], workflow['RequestName'])
            msg += " has been already subscribed to TAPE under request id: {}".format(dataItem['TapeRuleID'])
            self.logger.info(msg)
            return False

        dataTier = dataItem['Dataset'].split('/')[-1]
        if dataTier in self.msConfig['excludeDataTier']:
            msg = "Skipping tape data placement for dataset: {} ".format(dataItem['Dataset'])
            msg += "because it's been excluded in the MS configuration."
            self.logger.warning(msg)
            return False

        if dataTier in self.uConfig['tiers_with_no_custodial']['value']:
            msg = "Skipping tape data placement for dataset: {} ".format(dataItem['Dataset'])
            msg += "because Unified configuration sets it not to go to tape."
            self.logger.warning(msg)
            return False

        # if we are here, that means the dataset can proceed to tape
        return True

    def docUploader(self, msOutDoc, update=False, keys=None, stride=None):
        """
        A function to upload documents to MongoDB. The session object to the  relevant
        database and Collection must be passed as arguments
        :msOutDocs: A list of documents of type MSOutputTemplate
        :stride: the max number of documents we are about to upload at once
        :update: A flag to trigger document update in MongoDB in case of duplicates
        :keys:   A list of keys to update. If missing the whole document will be updated
        """
        # NOTE: We must return the document(s) at the end so that it can be explicitly
        #       deleted outside the pipeline

        # figure out which database collection to use, based on RelVal or standard workflow
        if msOutDoc["IsRelVal"]:
            dbColl = self.msOutRelValColl
        else:
            dbColl = self.msOutNonRelValColl
        # Skipping documents avoiding index unique property (documents having the
        # same value for the indexed key as an already uploaded document)
        try:
            dbColl.insert_one(msOutDoc)
        except errors.DuplicateKeyError:
            # DONE:
            #    Here we may wish to double check and make document update, so
            #    that a change of the Request on ReqMgr may be reflected here too
            # NOTE:
            #    If we use the 'update' option with a fresh document created from
            #    Reqmgr and we overwrite an already existing document in MongoDB
            #    which have been already worked on - we will loose the information
            #    that have been stored in the MonggDB - so always use 'update'
            #    with the proper set of keys to be updated
            if not keys:
                keys = []

            # update only the requested keys:
            if update and keys:
                updateDict = {}
                for key in keys:
                    updateDict[key] = msOutDoc[key]
                msOutDoc = dbColl.find_one_and_update(
                    {'_id': msOutDoc['_id']},
                    {'$set': updateDict},
                    return_document=ReturnDocument.AFTER)
            if update and not keys:
                msOutDoc = dbColl.find_one_and_update(
                    {'_id': msOutDoc['_id']},
                    {'$set': msOutDoc},
                    return_document=ReturnDocument.AFTER)
        return msOutDoc

    def getRecentDocNamesFromMongo(self):
        """
        Fetch a list of workflow names already inserted into MongoDB.
        :param dbColl: connection object to the database/collection
        :return: a flat list of workflow names
        """
        recordList = []
        # query for documents created after this timestamp
        createdAfter = int(time.time()) - self.msConfig['mongoDocsCreatedSecs']
        thisQuery = {"CreationTime": {"$gte": createdAfter}}
        projectionFields = {"RequestName": 1, "_id": 0}
        for dbColl in [self.msOutNonRelValColl, self.msOutRelValColl]:
            self.logger.info("Querying %s for docs created after timestamp: %s", dbColl.name, createdAfter)
            for mongoDoc in dbColl.find(thisQuery, projectionFields):
                recordList.append(mongoDoc["RequestName"])
        self.logger.info("Retrieved a total of %s recent requests from MongoDB", len(recordList))
        return recordList

    def getDocsFromMongo(self, mQueryDict, dbColl, limit=1000):
        """
        Reads documents from MongoDB and convert them to an MSOutputTemplate
        object. Limit can be provided to control the amount of records to be
        returned:
        :param mQueryDict: dictionary with the Mongo query to be executed
        :param dbColl: connection object to the database/collection
        :param limit: integer with the amount of documents meant to be returned
        :return: it yields an MSOutputTemplate object
        """
        # NOTE:
        # In case the current query returns an empty document from MongoDB
        # (eg. all workflows have been processed) the MSOutputTemplate
        # will throw an error. We should catch this one here and interrupt
        # the pipeLine traversal, otherwise an error either here or in one of the
        # following stages will most probably occur and the whole run will be broken.
        counter = 0
        for mongoDoc in dbColl.find(mQueryDict):
            if counter >= limit:
                return
            try:
                msOutDoc = MSOutputTemplate(mongoDoc, producerDoc=False)
                counter += 1
                yield msOutDoc
            except Exception as ex:
                msg = "Failed to create MSOutputTemplate object from mongo record: {}.".format(mongoDoc)
                msg += " Error message was: {}".format(str(ex))
                self.logger.exception(msg)
                raise ex
        if not counter:
            self.logger.info("%s Query: '%s' did not return any valid record from MongoDB", dbColl.name, mQueryDict)

    def getTransferInfo(self, reqName):
        """
        Searches and reads a document from MongoDB in all collections related to
        the MSOutput service. It is supposed to be called only by MSManager e.g.:
           transferDoc = self.msOutputProducer.getTransferInfo(reqName)
        And the output to be served to the REST interface, so that all request
        transfer records can be tracked.
        :param reqName: The name of the request to be searched for
        :return: a list of all msOutDocs with the record if it exists or None otherwise
        """

        mQueryDict = {'RequestName': reqName}
        result = []
        stripKeys = ['_id']

        for dbColl in [self.msOutNonRelValColl, self.msOutRelValColl]:
            doc = dbColl.find_one(mQueryDict)
            if doc:
                for key in stripKeys:
                    doc.pop(key, None)
                result.append(doc)
                break
        return result

    def docCleaner(self, doc):
        """
        Calls the dictionary internal method clear() and purges all the contents
        of the document
        """
        return doc.clear()

    def alertDIDNotFound(self, wflowName, containerList):
        """
        Send an alert to Prometheus for output containers not found within
        a given workflow.
        :param wflowName: string with the workflow name
        :param containerList: list of container names
        :return: none
        """
        alertName = "ms-output: output containers not found for workflow: {}".format(wflowName)
        alertSeverity = "high"
        alertSummary = "[MSOutput] Workflow '{}' has output datasets unknown to Rucio".format(wflowName)
        alertDescription = "Dataset(s): {} cannot be found in Rucio. ".format(containerList)
        alertDescription += "Thus, we are skipping these datasets from the final output "
        alertDescription += "data placement, such that this workflow can get archived."
        self.logger.warning(alertDescription)
        if self.msConfig["sendNotification"]:
            self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription, self.alertServiceName)

    def alertCampaignNotFound(self, campaignName, containerName):
        """
        Send an alert to Prometheus for campaign not found in the database.
        :param campaignName: string with the campaign name
        :param containerName: string with the container name
        :return: none
        """
        alertName = "ms-output: Campaign not found: {}".format(campaignName)
        alertSeverity = "high"
        alertSummary = "[MSOutput] Campaign '{}' not found in central CouchDB".format(campaignName)
        alertDescription = "Dataset: {} cannot have an output transfer rule ".format(containerName)
        alertDescription += "because its campaign: {} cannot be found in central CouchDB.".format(campaignName)
        alertDescription += " In order to get output data placement working, add it ASAP please."
        self.logger.critical(alertDescription)
        if self.msConfig["sendNotification"]:
            self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription, self.alertServiceName)

    def alertDatatierNotFound(self, datatierName, containerName, isRelVal):
        """
        Send an alert to Prometheus for datatier not found in the configuration.
        :param datatierName: string with the datatier name
        :param containerName: string with the container name
        :param isRelVal: boolean whether it's a RelVal workflow or not
        :return: none
        """
        alertName = "ms-output: Datatier not found: {}".format(datatierName)
        alertSeverity = "high"
        alertSummary = "[MSOutput] Datatier not found in the Unified configuration: {}".format(datatierName)
        alertDescription = "Dataset: {} contains a datatier: {}".format(containerName, datatierName)
        alertDescription += " not yet inserted into Unified configuration. "
        alertDescription += "Please add it ASAP. Letting it pass for now..."
        self.logger.critical(alertDescription)
        if self.msConfig["sendNotification"] and not isRelVal:
            self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription, self.alertServiceName)



#!/usr/bin/env python
"""
_AddIgnore_

MySQL implementation of DBSBufferFiles.AddIgnore
"""

from WMCore.Database.DBFormatter import DBFormatter


class AddIgnore(DBFormatter):
    sql = """INSERT IGNORE INTO dbsbuffer_file (lfn, dataset_algo, status, in_phedex)
                VALUES (:lfn, :dataset_algo, :status, :in_phedex)"""

    def execute(self, lfns, datasetAlgo, status, inPhEDEx,
                conn=None, transaction=False):
        binds = []
        for lfn in lfns:
            binds.append({"lfn": lfn, "dataset_algo": datasetAlgo,
                          "status": status, "in_phedex": inPhEDEx})

        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)
        return
